from .app_worker import DrSaiAPP

import os
from typing import (
    Union, 
    List, 
    Dict, 
    Any,
    AsyncGenerator,
    Mapping,
    )
from loguru import logger
import traceback
from dataclasses import dataclass, field
from fastapi import FastAPI
import uvicorn, asyncio
from autogen_agentchat.base import TaskResult
from autogen_agentchat.base import (
    ChatAgent, 
    Team)
from autogen_agentchat.ui import Console
from pathlib import Path

from hepai import HRModel, HModelConfig, HWorkerConfig, HWorkerAPP
import hepai

from drsai.modules.managers.database import DatabaseManager
from autogen_agentchat.base import (
    ChatAgent, 
    Team,)
from drsai.configs import CONST

from drsai.utils.utils import auto_worker_address, upload_to_hepai_filesystem

here = Path(__file__).parent.resolve()

############################################
# Dr.Sai application
############################################

async def start_console(
        task: str,
        agent_factory: callable = None, 
        agent: ChatAgent|Team = None, 
        init_kwargs: Dict = {},
        close_kwargs: Dict = {},
        ) -> Union[None, TaskResult]:
    """
    启动aotugen原生多智能体运行方式和多智能体逻辑
    args:
        task: str, 任务内容
        agent_factory: 工厂函数，用于创建AssistantAgent/BaseGroupChat实例
        agent: AssistantAgent|BaseGroupChat, 已创建的AssistantAgent/BaseGroupChat实例
        **kwargs: 其他参数
    """
    try:
        if agent is None:
            agent: ChatAgent|Team = (
                await agent_factory() 
                if asyncio.iscoroutinefunction(agent_factory)
                else agent_factory()
            )
        if hasattr(agent, "lazy_init"):
            await agent.lazy_init(**init_kwargs)
        stream = agent._model_client_stream if not isinstance(agent, Team) else agent._participants[0]._model_client_stream
        if stream:
            await Console(agent.run_stream(task=task))
            return 
        else:
            result:TaskResult = await agent.run(task=task)
            return result
    except Exception as e:
        logger.error(f"Error in a_drsai_ui_completions: {e}")
        traceback.print_exc()
    finally:
        if hasattr(agent, "close"):
            await agent.close(**close_kwargs)


async def run_console(agent_factory: callable, task: str, **kwargs) -> Union[None, TaskResult]:
    '''
    启动OpenAI-Style-API后端服务
    args:
        agent_factory: 工厂函数，用于创建AssistantAgent/BaseGroupChat实例
        task: str, 任务内容
        **kwargs: 其他参数
    '''

    result = await start_console(task=task, agent_factory=agent_factory, **kwargs)
    if result is not None:
        print(result)
        return result

async def run_backend(agent_factory: callable, **kwargs):
    '''
    启动OpenAI-Style-API后端服务
    args:
        agent_factory: 工厂函数，用于创建AssistantAgent/BaseGroupChat实例
        host: str = , "0.0.0.0" ,  # 后端服务host
        port: int = 42801,  # 后端服务port
        engine_uri: str = None,  # 数据库uri
        base_dir: str = None,  # 数据库目录
        auto_upgrade: bool = False,  # 是否自动升级数据库
        enable_openwebui_pipeline: bool = False,  # 是否启动openwebui pipelines
        agnet_name: str = "Dr.Sai",  # 智能体的名称
        pipelines_dir: str = None,  # openwebui pipelines目录
        history_mode: str = "backend",  # 历史消息的加载模式，可选值：backend、frontend 默认backend
        use_api_key_mode: str = "frontend",  # api key的使用模式，可选值：frontend、backend 默认frontend， 调试模式下建议设置为backend
    '''
    host: str =  kwargs.pop("host", "0.0.0.0")
    port: int =  kwargs.pop("port", 42801)
    engine_uri = kwargs.pop('engine_uri', None) or f"sqlite:///{CONST.FS_DIR}/drsai.db"
    base_dir = kwargs.pop('base_dir', None) or CONST.FS_DIR
    db_manager = DatabaseManager(
        engine_uri = engine_uri,
        base_dir = base_dir
    )
    auto_upgrade = kwargs.pop('auto_upgrade', False)
    init_response = db_manager.initialize_database(auto_upgrade=auto_upgrade)
    assert init_response.status, init_response.message
    kwargs.update({"db_manager": db_manager})

    enable_pipeline: bool = kwargs.pop("enable_openwebui_pipeline", False)
    drsaiapp = DrSaiAPP(
        agent_factory = agent_factory,
        **kwargs
        )
    if enable_pipeline:
        os.environ['BACKEND_PORT'] = str(port)
        agnet_name = kwargs.pop("agnet_name", "Dr.Sai")
        os.environ['AGNET_NAME'] = agnet_name
        pipelines_dir = kwargs.pop("pipelines_dir", None)
        if pipelines_dir is not None:
            os.environ['PIPELINES_DIR'] = pipelines_dir
            pipelines_dir = os.getenv('PIPELINES_DIR')
            if not os.path.exists(pipelines_dir):
                print(f"PIPELINES_DIR {pipelines_dir} not exists!")
            else:
                print(f"Set PIPELINES_DIR to {pipelines_dir}")

        from contextlib import asynccontextmanager
        # 通过Pipeline适配OpenWebUI
        from .owebui_pipeline.api import app as owebui_pipeline_app
        from .owebui_pipeline.api import lifespan as owebui_lifespan
        
        drsaiapp.app.mount("/pipelines", app=owebui_pipeline_app)
        main_lifespan = getattr(drsaiapp.app.router, "lifespan_context", None)

        # 创建组合生命周期上下文
        @asynccontextmanager
        async def combined_lifespan(app: FastAPI):
            # 执行主应用初始化（如果存在）
            if main_lifespan:
                async with main_lifespan(app):
                    # 执行子应用生命周期
                    async with owebui_lifespan(app):
                        yield
            else:
                # 仅执行子应用生命周期
                async with owebui_lifespan(app):
                    yield
        
        # 重写主应用生命周期
        drsaiapp.app.router.lifespan_context = combined_lifespan

    config = uvicorn.Config(
        app=drsaiapp.app,
        host=host,
        port=port,
        loop="asyncio"
    )
    server = uvicorn.Server(config)
    # 在现有事件循环中启动服务
    if enable_pipeline:
        print(f"Enable OpenWebUI pipelines: `http://{host}:{port}/pipelines` with API-KEY: `{owebui_pipeline_app.api_key}`")
    try:
        await server.serve()
    except asyncio.CancelledError:
        await server.shutdown()
    finally:
        # 关闭数据库连接
        await db_manager.close()


@dataclass
class DrSaiModelConfig(HModelConfig):
    name: str = field(default="drsai/besiii", metadata={"help": "Model's name"})
    permission: Union[str, Dict] = field(default=None, metadata={"help": "Model's permission, separated by ;, e.g., 'groups: all; users: a, b; owner: c', will inherit from worker permissions if not setted"})
    version: str = field(default="2.0", metadata={"help": "Model's version"})
    
@dataclass
class DrSaiWorkerConfig(HWorkerConfig):
    host: str = field(default="0.0.0.0", metadata={"help": "Worker's address, enable to access from outside if set to `0.0.0.0`, otherwise only localhost can access"})
    port: int = field(default=42801, metadata={"help": "Worker's port, default is None, which means auto start from `auto_start_port`"})
    auto_start_port: int = field(default=42801, metadata={"help": "Worker's start port, only used when port is set to `auto`"})
    route_prefix: str = field(default="/apiv2", metadata={"help": "Route prefix for worker"})
    # controller_address: str = field(default="https://aiapi001.ihep.ac.cn", metadata={"help": "The address of controller"})
    controller_address: str = field(default="http://localhost:42601", metadata={"help": "The address of controller"})
    
    controller_prefix: str = field(default="/apiv2", metadata={"help": "Controller's route prefix"})
    no_register: bool = field(default=True, metadata={"help": "Do not register to controller"})
    

    permissions: str = field(default='groups: payg; users: admin, xiongdb@ihep.ac.cn, ddf_free; owner: xiongdb@ihep.ac.cn', metadata={"help": "Model's permissions, separated by ;, e.g., 'groups: default; users: a, b; owner: c'"})
    description: str = field(default='This is Dr.Sai multi agents system', metadata={"help": "Model's description"})
    author: str = field(default=None, metadata={"help": "Model's author"})
    daemon: bool = field(default=False, metadata={"help": "Run as daemon"})
    type: str = field(default="agent", metadata={"help": "Worker's type"})
    debug: bool = field(default=False, metadata={"help": "Debug mode"})
    _metadata: dict = field(default_factory=dict, metadata={"help": "Additional metadata for worker/model"})


class DrSaiWorkerModel(HRModel):  # Define a custom worker model inheriting from HRModel.
    def __init__(
            self, 
            config: DrSaiModelConfig,
            worker_config: DrSaiWorkerConfig,
            logo: str = "https://aiapi.ihep.ac.cn/apiv2/files/file-8572b27d093f4e15913bebfac3645e20/preview",
            examples: List[str] = [],
            agent_config: Dict[str, Any] = {},
            defult_config_name: str|None = None,
            close_agent_on_finish: bool = True,
            drsaiapp: DrSaiAPP = None # 传入DrSaiAPP实例
            ):
        super().__init__(config=config)

        # if drsaiapp is not None and isinstance(drsaiapp, type):
        #     self.drsai = drsaiapp()  # Instantiate the DrSaiAPP instance.
        # else:
        #     self.drsai = drsaiapp or DrSaiAPP()  # Instantiate the DrSaiAPP instance.
        # pass
        self.drsai: DrSaiAPP = drsaiapp
        self._info = {
            "name": config.name, 
            "description": worker_config.description, 
            "version": config.version, 
            "author": worker_config.author, 
            "logo": logo,
            "examples": examples,
            "agent_config": agent_config,
            "defult_config_name": defult_config_name,
            } 
        self.drsai._info = self._info

        self._close_agent_on_finish = close_agent_on_finish

        
    @HRModel.remote_callable
    async def get_info(self) -> Dict[str, str]:
       return self._info
    
    @HRModel.remote_callable
    async def a_get_agents_info(self, chat_id: str) -> List[Dict[str, Any]]:
        agent: Team|ChatAgent = self.drsai.agent_instance.get(chat_id, None)
        return await self.drsai.get_agents_info(agent=agent)
    
    @HRModel.remote_callable
    async def lazy_init(self, chat_id: str, api_key: str, run_info: Dict[str, str], stream: bool = True, **kwargs) -> Dict[str, Any]:
        try:
            agent: Team|ChatAgent = self.drsai.agent_instance.get(chat_id, None)
            if agent is None:
                agent = await self.drsai._create_agent_instance(
                    api_key=api_key, 
                    thread_id=chat_id, 
                    user_id=run_info.get("email"), 
                    stream=stream,
                    defult_config_name=kwargs.get("defult_config_name", None),
                    )
            message = await agent.lazy_init(api_key=api_key, thread_id=chat_id, run_info=run_info, **kwargs)
            return {"status": True, "message": message}
        except Exception as e:
            return {"status": False, "message": f"Lazy init error: {e}"}
    
    @HRModel.remote_callable
    async def pause(self, chat_id: str) -> Dict[str, Any]:
        try:
            agent: Team|ChatAgent = self.drsai.agent_instance[chat_id]
            message = await agent.pause()
            if message is None:
                message = ""
            return {"status": True, "message": message}
        except Exception as e:
            return {"status": False, "message": f"Pause error: {e}"}
    
    @HRModel.remote_callable
    async def pause_long_task(self, chat_id: str) -> Dict[str, Any]:
        try:
            agent: Team|ChatAgent = self.drsai.agent_instance[chat_id]
            if hasattr(agent, "long_task_pause"):
                await agent.long_task_pause()  # type: ignore
            return {"status": True, "message": ""}
        except Exception as e:
            return {"status": False, "message": f"Pause long task error: {e}"}
    
    @HRModel.remote_callable
    async def resume(self, chat_id: str) -> Dict[str, Any]:
        try:
            agent: Team|ChatAgent = self.drsai.agent_instance[chat_id]
            await agent.resume()
            return {"status": True, "message": ""}
        except Exception as e:
            return {"status": False, "message": f"Resume error: {e}"}
    
    @HRModel.remote_callable
    async def close(self, chat_id: str) -> Dict[str, Any]:
        try:
            agent: Team|ChatAgent = self.drsai.agent_instance[chat_id]
            if self._close_agent_on_finish:
                await agent.close()
                self.drsai.agent_instance.pop(chat_id, None)
            return {"status": True, "message": ""}
        except Exception as e:
            return {"status": False, "message": f"Close error: {e}"}
        
    @HRModel.remote_callable
    async def save_state(self, chat_id: str) -> Dict[str, Any]:
        try:
            agent: Team|ChatAgent = self.drsai.agent_instance[chat_id]
            await agent.save_state()
            return {"status": True, "message": ""}
        except Exception as e:
            return {"status": False, "message": f"save_state error: {e}"}
    
    @HRModel.remote_callable
    async def load_state(self, chat_id: str, state: Dict[str, Any]) -> Dict[str, Any]:
        try:
            agent: Team|ChatAgent = self.drsai.agent_instance[chat_id]
            await agent.load_state(state=state)
            return {"status": True, "message": ""}
        except Exception as e:
            return {"status": False, "message": f"load_state error: {e}"}

    @HRModel.remote_callable
    async def a_chat_completions(self, *args, **kwargs) -> AsyncGenerator:
        return self.drsai.a_drsai_ui_completions(*args, **kwargs)
    
    @HRModel.remote_callable
    async def chat_completions(self, *args, **kwargs) -> AsyncGenerator:
        return self.drsai.a_start_chat_completions(*args, **kwargs)


async def run_worker(agent_factory: callable, **kwargs):
    '''
    启动HepAI-Worker-Style-API后端服务
    args:
        agent_factory: 工厂函数，用于创建AssistantAgent/BaseGroupChat实例
        agent_name: str = , "Dr.Sai" ,  # 智能体的名称
        description: str = , "Dr.Sai is a helpful assistant." ,  # 智能体的描述
        examples: List[str] = [],  # 智能体的示例
        llm_mode_config: Dict[str, Any] = {},# LLM模式配置
        version: str = , "0.1.0" ,  # 智能体的版本
        logo: str = , "https://aiapi.ihep.ac.cn/apiv2/files/file-8572b27d093f4e15913bebfac3645e20/preview" ,  # 智能体的logo
        host: str = , "0.0.0.0" ,  # 后端服务host
        port: int = 42801,  # 后端服务port
        no_register: bool = False,  # 是否注册到控制器
        controller_address: str = "https://aiapi.ihep.ac.cn",  # 控制器地址
        engine_uri: str = "sqlite:///~/drsai/drsai.db" # 智能体数据库地址
        base_dir: str = "~/drsai", # 数据库目录
        history_mode: str = "backend",  # 历史消息的加载模式，可选值：backend、frontend 默认backend
        use_api_key_mode: str = "frontend",  # api key的使用模式，可选值：frontend、backend 默认frontend， 调试模式下建议设置为backend
        enable_pipeline: bool = False,  # 是否启动openwebui pipelines
        join_topics: List[str] = [],  # 是否为智能体添加默认的join_topics
        close_kwargs: 
    '''
    model_args_obj: DrSaiModelConfig = DrSaiModelConfig
    worker_args_obj: DrSaiWorkerConfig = DrSaiWorkerConfig
    model_args, worker_args = hepai.parse_args((model_args_obj, worker_args_obj))

    agent_name: str = kwargs.pop("agent_name", None)
    if agent_name is not None:
        model_args.name = agent_name
        os.environ['AGNET_NAME'] = agent_name
    
    author: str = kwargs.pop("author", "IsYourBaby")
    worker_args.author = author

    permission: str|dict = kwargs.pop("permission", None)
    if permission is not None:
        if isinstance(permission, dict):
            groups = "groups: " + permission.get('groups', "default")
            users = "users: " + ", ".join(permission.get('users', []))
            owner = "owner: " + permission.get('owner', "")
            worker_args.permissions = "; ".join([groups, users, owner])
        else:
            worker_args.permissions = permission
    
    description: str = kwargs.pop("description", "A Dr.Sai multi agents system")
    worker_args.description = description

    version: str = kwargs.pop("version", None)
    if version is not None:
        model_args.version = version
    
    logo: str = kwargs.pop("logo", 'https://aiapi.ihep.ac.cn/apiv2/files/file-8572b27d093f4e15913bebfac3645e20/preview')
    if os.path.exists(logo):
        logo_path = Path(logo)
        if logo_path.is_file():
            file_obj = upload_to_hepai_filesystem(str(logo_path))
            logo = file_obj["url"]
    examples: List[str] = kwargs.pop("examples", [])
    agent_config: Dict[str, Any] = kwargs.pop("agent_config", {})
    defult_config_name: str|None = kwargs.pop("defult_config_name", None)
    
    host: str =  kwargs.pop("host", None)
    if host is not None:
        worker_args.host = host

    port: int =  kwargs.pop("port", None)
    if port is not None:
        worker_args.port = port
        os.environ['BACKEND_PORT'] = str(port)
    
    drsai_dir = kwargs.pop('drsai_dir', None) or CONST.FS_DIR
    engine_uri = kwargs.pop('engine_uri', None) or f"sqlite:///{drsai_dir}/drsai.db"
    base_dir = kwargs.pop('base_dir', None) or drsai_dir
    db_manager = DatabaseManager(
        engine_uri = engine_uri,
        base_dir = base_dir
    )
    auto_upgrade = kwargs.pop('auto_upgrade', False)
    init_response = db_manager.initialize_database(auto_upgrade=auto_upgrade)
    assert init_response.status, init_response.message
    kwargs.update({"db_manager": db_manager})

    no_register: bool =  kwargs.pop("no_register", None)
    if no_register is not None:
        worker_args.no_register = no_register

    controller_address: str =  kwargs.pop("controller_address", "https://aiapi.ihep.ac.cn")
    worker_args.controller_address = controller_address

    close_agent_on_finish: bool = kwargs.pop("close_agent_on_finish", True)

    # TODO: ADD METADATA for worker config
    _metadata: dict[str, Any] = kwargs.pop("metadata", None)
    if _metadata is not None:
       worker_args._metadata.update(_metadata)

    join_topics: List[str]|None = kwargs.pop("join_topics", None)
    if join_topics is not None:
        worker_args._metadata.update({"join_topics": join_topics})
    
    close_kwargs: dict[str, Any] = kwargs.pop("close_kwargs", {})

    # WeChat
    link_wechat = kwargs.pop("link_wechat", False)

    print(model_args)
    print()
    print(worker_args)
    print()

    drsaiapp = DrSaiAPP(
        agent_factory = agent_factory,
        **kwargs
        )
    
    model = DrSaiWorkerModel(
        config=model_args, 
        worker_config=worker_args, 
        logo=logo, 
        examples=examples,
        agent_config = agent_config,
        defult_config_name = defult_config_name,
        close_agent_on_finish = close_agent_on_finish,
        drsaiapp=drsaiapp)

    # ── WeChat Bot 集成 ────────────────────────────────────────────────────────
    _wechat_tasks: list[asyncio.Task] = []
    if link_wechat:
        from drsai.configs.constant import WECHAT_DIR
        from .wechat.wechat_login import login_wechat_main, load_credentials
        from .wechat.session_manager import SessionManager
        from .wechat.wechat_bot import WeChatBot
        from .wechat.idle_monitor import idle_monitor

        wechat_dir = WECHAT_DIR
        if base_dir != CONST.FS_DIR:
            wechat_dir  = os.path.join(base_dir, "wechat")
            os.makedirs(wechat_dir, exist_ok=True)
        creds_file = os.path.join(wechat_dir, "credentials.json")

        # 1. QR 扫码登录 + HEPAI_API_KEY 录入（交互式，阻塞直到扫码完成）
        await login_wechat_main(creds_file = creds_file)
        creds = load_credentials(creds_file = creds_file)
        api_key = creds.get("hepai_api_key") or os.environ.get("HEPAI_API_KEY", "")

        # 2. 初始化会话管理器和 Bot
        
        sessions_file = os.path.join(wechat_dir, "sessions.json")
        session_mgr = SessionManager(sessions_file)
        bot = WeChatBot(
            model=model,
            creds=creds,
            api_key=api_key,
            session_manager=session_mgr,
        )

        # 3. 启动后台任务（与 uvicorn 共享同一 event loop）
        _wechat_tasks.append(asyncio.create_task(bot.run(), name="wechat_bot"))
        _wechat_tasks.append(asyncio.create_task(
            idle_monitor(model, session_mgr), name="wechat_idle_monitor"
        ))
        print("微信 Bot 已启动，发送 /help 查看命令列表。")

    enable_pipeline: bool = kwargs.pop("enable_openwebui_pipeline", False)
    if enable_pipeline:
        pipelines_dir = kwargs.pop("pipelines_dir", None)
        if pipelines_dir is not None:
            os.environ['PIPELINES_DIR'] = pipelines_dir
            pipelines_dir = os.getenv('PIPELINES_DIR')
            if not os.path.exists(pipelines_dir):
                print(f"PIPELINES_DIR {pipelines_dir} not exists!")
            else:
                print(f"Set PIPELINES_DIR to {pipelines_dir}")
        # 通过Pipeline适配OpenWebUI
        from .owebui_pipeline.api import app as owebui_pipeline_app
        from .owebui_pipeline.api import lifespan as owebui_lifespan
        # 实例化HWorkerAPP
        app: FastAPI = HWorkerAPP(
            model, worker_config=worker_args,
            lifespan=owebui_lifespan,
            )  # Instantiate the APP, which is a FastAPI application.
        app.mount("/pipelines", app=owebui_pipeline_app)
        
    else:
        app: FastAPI = HWorkerAPP(
            model, worker_config=worker_args
            )
    
    app.include_router(model.drsai.router)
    
    print(app.worker.get_worker_info(), flush=True)
    # # 启动服务
    # uvicorn.run(self.app, host=self.app.host, port=self.app.port)
    # 创建uvicorn配置和服务实例
    config = uvicorn.Config(
        app, 
        host=worker_args.host,  # 确保这里使用的是正确的host参数
        port=worker_args.port   # 确保这里使用的是正确的port参数
    )
    server = uvicorn.Server(config)
    # 在现有事件循环中启动服务
    worker_address = auto_worker_address(worker_address='auto', host=worker_args.host, port=worker_args.port)
    print(f"#####################Your Agent Server is ready!######################")
    print(f"Enable HepAI worker URL: `{worker_address}/apiv2`")
    if enable_pipeline:
        print(f"Enable OpenWebUI pipelines URL: `{worker_address}/pipelines` with API-KEY: `{owebui_pipeline_app.api_key}`")
    print(f"#####################################################################")
    try:
        await server.serve()
    finally:
        # 关闭微信后台任务
        for task in _wechat_tasks:
            task.cancel()
        if _wechat_tasks:
            await asyncio.gather(*_wechat_tasks, return_exceptions=True)
        # 关闭数据库连接
        await db_manager.close()
        for agent in model.drsai.agent_instance:
            if hasattr(agent, "close"):
                await agent.close(**close_kwargs)






__version__ = "1.0.7"
__appname__ = 'drsai'
__author__ = 'hepai'
__email__ = 'xiongdb@ihep.ac.cn/hepai@ihep.ac.cn'
__description__ = "Developed by Dr. Sai's team at the Institute of High Energy Physics, Chinese Academy of Sciences (IHEP, CAS), this is a development framework designed for both single-agent and multi-agent collaborative systems"


from autogen_core._component_config import (
    Component,
    ComponentBase,
    ComponentFromConfig,
    ComponentLoader,
    ComponentModel,
    ComponentSchemaType,
    ComponentToConfig,
    ComponentType,
    is_component_class,
    is_component_instance,
)



from typing import (
    AsyncGenerator, 
    List, 
    Sequence, 
    Dict, 
    Any, 
    Callable, 
    Awaitable, 
    Union, 
    Optional, 
    Tuple,
    Self,
    Mapping,
    )

import asyncio, traceback, json
from pydantic import BaseModel
from pathlib import Path
from loguru import logger

from drsai import CancellationToken, FunctionCall
from drsai.modules.baseagent import (
    DrSaiAgent, 
    HandoffBase, 
    Response, 
    TaskResult, 
    CreateResult,
    FunctionExecutionResultMessage,
    AssistantMessage,
    UserMessage,
    SystemMessage,
    LLMMessage
    )
from drsai.modules.baseagent.drsaiagent import DrSaiAgentConfig
from drsai.modules.baseagent import CodeExecutorAgent, CodeExecutor
from drsai.modules.components import (
    ComponentModel,
)
from drsai.modules.components.model_client import ChatCompletionClient
from drsai.modules.components.model_context import (
    ChatCompletionContext,
)
from drsai.modules.components.memory import Memory
from drsai.modules.components.tool import (
    BaseTool, 
    FunctionTool, 
    Workbench,
    ToolSchema,
    ParametersSchema,
    )
from drsai.modules.managers.messages import (
    BaseAgentEvent,
    BaseChatMessage,
    AgentEvent,
    ChatMessage,
    HandoffMessage,
    MemoryQueryEvent,
    ModelClientStreamingChunkEvent,
    TextMessage,
    StopMessage,
    ToolCallExecutionEvent,
    ToolCallRequestEvent,
    ToolCallSummaryMessage,
    UserInputRequestedEvent,
    ThoughtEvent,
    AgentLogEvent,
    
    StructuredMessage,
    StructuredMessageFactory,
    MultiModalMessage,
    Image,
)
from drsai.modules.managers.database import DatabaseManager
from drsai.modules.components.skills import SkillLoader
from .managers.operater_funs import get_operator_funcs
from .managers.todo_manager import TodoManager
from .managers.user_profile_manager import UserProfileManager
from .managers.task_planner import TaskPlanner, TaskPlan
from .managers.memory_manager import LongTermMemoryManager
from drsai.configs.constant import RUNS_DIR

class EdgeAgentConfig(DrSaiAgentConfig):
    skills_dir: str
    executor: ComponentModel
    work_dir: str
    sub_agent_config: Dict | None

class EdgeAgent(DrSaiAgent):
    """
    EdgeAgent - 专业科学数据智能分析智能体

    核心能力:
    1. 任务规划与分解
    2. 多任务进度管理
    3. 智能工具/Skills/子智能体调用
    4. 长期记忆与学习
    5. 错误处理与用户交互
    """

    def __init__(
        self,
        name: str,
        *,
        model_client: ChatCompletionClient = None,
        tools: List[BaseTool[Any, Any] | Callable[..., Any] | Callable[..., Awaitable[Any]]] | None = None,
        workbench: Workbench | None = None,
        handoffs: List[HandoffBase | str] | None = None,
        model_context: ChatCompletionContext | None = None,
        description: str = "An agent that provides assistance with ability to use tools.",
        system_message: (
            str | None
        ) = "You are a helpful AI assistant. Solve tasks using your tools. Reply with TERMINATE when the task has been completed.",
        model_client_stream: bool = True,
        reflect_on_tool_use: bool | None = None,
        tool_call_summary_format: str = "{result}",
        tool_call_summary_prompt: str| None = None,
        output_content_type: type[BaseModel] | None = None,
        output_content_type_format: str | None = None,
        memory: Sequence[Memory] | None = None,
        metadata: Dict[str, str] | None = None,

        # drsaiAgent specific
        skills_dir: str | None = None,
        skills_loader: SkillLoader | None = None,
        work_dir: str | None = None,
        executor: CodeExecutor | None = None,
        sub_agent_config: Dict | None = None,
        max_turn_count: int = 20,
        memory_function: Callable | None = None,
        reply_function: Callable | None = None,
        db_manager: DatabaseManager | None = None,
        thread_id: str | None = None,
        user_id: str | None = None,
    ):
        super().__init__(
            name=name,
            model_client=model_client,
            tools=tools,
            workbench=workbench,
            handoffs=handoffs,
            model_context=model_context,
            description=description,
            system_message=system_message,
            model_client_stream=model_client_stream,
            reflect_on_tool_use=reflect_on_tool_use,
            tool_call_summary_format=tool_call_summary_format,
            tool_call_summary_prompt=tool_call_summary_prompt,
            output_content_type=output_content_type,
            output_content_type_format=output_content_type_format,
            memory=memory,
            metadata=metadata,
            memory_function=memory_function,
            reply_function=reply_function,
            db_manager=db_manager,
            thread_id=thread_id,
            user_id=user_id,
        )

        if not work_dir:
            if self._db_manager:
                DEFAULT_RUN_DIR: Path = self._db_manager.schema_manager.base_dir / "runs"
                self._work_dir = DEFAULT_RUN_DIR
            else:
                self._work_dir = Path(RUNS_DIR)
        else:
            self._work_dir = Path(work_dir)
        if not self._work_dir.exists():
            self._work_dir.mkdir(parents=True)

        # === initial UserProfileManager ===
        self._user_profile_manager = UserProfileManager(
            work_dir=self._work_dir,
            user_id=self._user_id,
        )

        # 获取用户画像并更新系统消息
        user_context = self._user_profile_manager.get_agent_system_prompt()
        if user_context and system_message:
            # 合并系统消息和用户上下文
            enhanced_system_message = f"""{system_message}

{user_context}
"""
            self._system_messages = [SystemMessage(content=enhanced_system_message)]

        # basic tools
        self._basic_funcs: List[Callable] = get_operator_funcs(work_dir)
        self._basic_funcs_names = [func.__name__ for func in self._basic_funcs]
        for func in self._basic_funcs:
            self._tools.append(FunctionTool(func, description=func.__doc__))
        self._workbench._tools = self._tools

        # skills - 优先从用户目录加载
        if skills_loader:
            self._skills_loader = skills_loader
        else:
            # 首先尝试从用户的skills目录加载
            user_skills_dir = self._user_profile_manager.skills_dir
            if user_skills_dir.exists() and list(user_skills_dir.glob("*/SKILL.md")):
                self._skills_loader = SkillLoader(skills_dir=str(user_skills_dir))
            elif skills_dir:
                self._skills_loader = SkillLoader(skills_dir=skills_dir)
            else:
                # 创建空的skills loader
                user_skills_dir.mkdir(parents=True, exist_ok=True)
                self._skills_loader = SkillLoader(skills_dir=str(user_skills_dir))

        if self._skills_loader.skills:
            self._agent_skills_tools = [self.get_agent_skills_tools()]
        else:
            self._agent_skills_tools = []

        # executor
        self._executor = executor
        
        # sub_agent_config
        self._sub_agent_config = sub_agent_config
        if self._sub_agent_config:
            self._subagent_tools = []
        else:
            self._subagent_descriptions  = self.get_subagent_descriptions()
            self._subagent_tools = [self.get_subagent_tools()]
        

        # max_turn_count
        self._max_turn_count = max_turn_count

        #  todo manager
        self._todo_manager = TodoManager()
        self._todo_tools = [self.get_todo_manager_tools()]

        # 初始化实例变量供edge_agent_core使用
        self._current_plan = None
        self._task_planner = None
        self._memory_manager = None

    async def on_messages_stream(
        self, messages: Sequence[BaseChatMessage], cancellation_token: CancellationToken
    ) -> AsyncGenerator[BaseAgentEvent | BaseChatMessage | Response, None]:
        """
        EdgeAgent的主消息处理循环
        使用EdgeAgentCore的完整任务规划和执行逻辑
        """
        # monitor the pause event
        if self.is_paused:
            yield Response(
                chat_message=TextMessage(
                    content=f"The {self.name} is paused.",
                    source=self.name,
                    metadata={"internal": "yes"},
                )
            )
            return

        # Set up background task to monitor the pause event and cancel the task if paused.
        async def monitor_pause() -> None:
            await self._paused.wait()
            self.is_paused = True

        monitor_pause_task = asyncio.create_task(monitor_pause())
        inner_messages: List[BaseAgentEvent | BaseChatMessage] = []
        try:
            # Gather all relevant state here
            agent_name = self.name
            model_context = self._model_context
            memory = self._memory
            system_messages = self._system_messages
            workbench = self._workbench
            handoff_tools = self._handoff_tools
            handoffs = self._handoffs
            manager_tools = self._agent_skills_tools+self._todo_tools+self._subagent_tools
            model_client = self._model_client
            model_client_stream = self._model_client_stream
            reflect_on_tool_use = self._reflect_on_tool_use
            tool_call_summary_format = self._tool_call_summary_format
            output_content_type = self._output_content_type
            format_string = self._output_content_type_format
            
            # Add new user/handoff messages to the model context
            await self._add_messages_to_context(
                model_context=model_context,
                messages=messages,
            )

            # 获取最后一条消息
            last_message = messages[-1]
            logger.info(f"Received message from {last_message.source}")

            # === 阶段1: 判断消息来源和类型 ===
            if last_message.source == "user":
                # 首次用户请求 → 任务规划阶段
                yield await self._handle_initial_user_request(
                    last_message=last_message,
                    agent_name=agent_name,
                    model_context=model_context,
                    cancellation_token=cancellation_token,
                    inner_messages=inner_messages,
                )
                return

            elif last_message.source == "user_proxy":
                # 用户对计划的反馈 → 执行或重新规划
                async for event in self._handle_user_feedback(
                    last_message=last_message,
                    agent_name=agent_name,
                    model_context=model_context,
                    model_client=model_client,
                    model_client_stream=model_client_stream,
                    cancellation_token=cancellation_token,
                    inner_messages=inner_messages,
                ):
                    if self.is_paused:
                        raise asyncio.CancelledError()
                    yield event
                return

        except asyncio.CancelledError:
            # If the task is cancelled, we respond with a message.
            yield Response(
                chat_message=TextMessage(
                    content="The task was cancelled by the user.",
                    source=self.name,
                    metadata={"internal": "yes"},
                ),
                inner_messages=inner_messages,
            )
        except Exception as e:
            # logger.error(f"Error in {self.name}: {e}")
            logger.exception(f"Error in {self.name}")
            # add to chat history
            await model_context.add_message(
                AssistantMessage(
                    content=f"An error occurred while executing the task: {e}",
                    source=self.name
                )
            )
            yield Response(
                chat_message=TextMessage(
                    content=f"An error occurred while executing the task: {e}",
                    source=self.name,
                    metadata={"internal": "no"},
                ),
                inner_messages=inner_messages,
            )
        finally:
            # Cancel the monitor task.
            try:
                monitor_pause_task.cancel()
                await monitor_pause_task
            except asyncio.CancelledError:
                pass
    
    def get_todo_manager_tools(self, strict: bool = False,) -> ToolSchema:
        parameters = ParametersSchema(
            type="object",
            properties={
                "items": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "content": {"type": "string"},
                            "status": {
                                "type": "string",
                                "enum": ["pending", "in_progress", "completed"]
                            },
                            # "activeForm": {"type": "string"},
                        },
                    },
                },
            },
            required=["content", "status"], # , "activeForm"
            additionalProperties=False,
        )
        tool_schema = ToolSchema(
            name="TodoWrite",
            description="Create/Update task list.",
            parameters=parameters,
            strict=strict,
        )
        return tool_schema
    
    def get_subagent_tools(self, strict: bool = False,) -> ToolSchema:
        parameters = ParametersSchema(
            type="object",
            properties={
                "description": {
                    "type": "string",
                    "description":  "Short task description (3-5 words)"
                },
                "prompt": {
                    "type": "string",
                    "description": "The specific tasks that need to be executed by the sub agent. If the tasks include code blocks, files, etc. that need to be executed, they must be filled in completely."
                },
                "agent_type": {
                    "type": "string",
                    "enum": list(self._sub_agent_config.keys())
                },
            },
            required=["description", "prompt", "agent_type"],
            additionalProperties=False,
        )
        tool_schema = ToolSchema(
            name="Task",
            description=f"Spawn a subagent for a focused subtask.\n\nAgent types:\n{self._subagent_descriptions}",
            parameters=parameters,
            strict=strict,
        )
        return tool_schema
    
    def get_subagent_descriptions(self) -> str:
        """Generate agent type descriptions for system prompt."""
        return "\n".join(
            f"- {name}: {cfg['description']}"
            for name, cfg in self._sub_agent_config.items()
        )

    async def _call_llm(
        self,
        model_client: ChatCompletionClient,
        model_client_stream: bool,
        system_messages: List[SystemMessage],
        model_context: ChatCompletionContext,
        workbench: Workbench,
        handoff_tools: List[BaseTool[Any, Any]],
        manager_tools: List[ToolSchema],
        agent_name: str,
        cancellation_token: CancellationToken,
        output_content_type: type[BaseModel] | None,
    ) -> AsyncGenerator[Union[CreateResult, ModelClientStreamingChunkEvent], None]:
        """
        Perform a model inference and yield either streaming chunk events or the final CreateResult.
        """
        all_messages = await model_context.get_messages()
        
        llm_messages: List[LLMMessage] = self._get_compatible_context(model_client=model_client, messages=system_messages + all_messages)

        # 自定义的memory_function，用于RAG检索等功能，为大模型回复增加最新的知识
        if self._memory_function is not None:
            llm_messages = await self._call_memory_function(llm_messages, model_client, cancellation_token, agent_name)

        all_tools = (await workbench.list_tools()) + handoff_tools + manager_tools
        # model_result: Optional[CreateResult] = None
        if self._reply_function is not None:
            # 自定义的reply_function，用于自定义对话回复的定制
            async for chunk in self._call_reply_function(
                llm_messages, 
                model_client = model_client, 
                workbench=workbench,
                handoff_tools=handoff_tools,
                tools = all_tools,
                agent_name=agent_name, 
                cancellation_token=cancellation_token,
                db_manager=self._db_manager,
            ):
                # if isinstance(chunk, CreateResult):
                #     model_result = chunk
                yield chunk
        else:
           async for chunk in self.call_llm(
                agent_name = agent_name,
                model_client = model_client,
                llm_messages = llm_messages, 
                tools = all_tools, 
                model_client_stream = model_client_stream,
                cancellation_token = cancellation_token,
                output_content_type = output_content_type,
           ):
               yield chunk

    async def _handle_initial_user_request(
        self,
        last_message: BaseChatMessage,
        agent_name: str,
        model_context,
        cancellation_token: CancellationToken,
        inner_messages: List,
    ) -> Response:
        """
        处理用户的初始请求 - 任务规划阶段
        """
        # 获取用户请求
        user_request = last_message.content
        if hasattr(last_message, 'metadata'):
            user_request = last_message.metadata.get("user_request", user_request)

        # logger.info(f"Handling initial user request: {user_request[:100]}...")

        # 初始化任务规划器
        task_planner = TaskPlanner(
            agent_name=agent_name,
            model_client=self._model_client,
            skills_description=self._skills_loader.get_descriptions() if hasattr(self, '_skills_loader') else None,
            enable_skills_for_planning=True,
        )

        # 初始化记忆管理器
        memory_manager = LongTermMemoryManager(
            user_profile_manager=self._user_profile_manager,
            model_context=model_context,
            session_id=self._thread_id,
        )

        # 保存到实例变量供后续使用
        self._task_planner = task_planner
        self._memory_manager = memory_manager

        # 开始会话记录
        memory_manager.start_session(user_request=user_request, needs_plan=True)

        # 检索相关历史记忆(可选,用于辅助规划)
        # await memory_manager.inject_relevant_memories_to_context(
        #     keywords=self._extract_keywords_from_request(user_request),
        #     limit=2
        # )

        # 生成任务规划
        task_plan: TaskPlan = await task_planner.analyze_and_plan(
            user_request=user_request,
            model_context=model_context,
            cancellation_token=cancellation_token,
            use_skills_context=True,  # 使用skills上下文辅助规划
        )

        # 保存计划到记忆
        memory_manager.set_task_plan(task_plan.model_dump())

        # 发送计划给用户
        plan_message = await task_planner.send_plan_to_user(
            task_plan=task_plan,
            agent_name=agent_name,
        )

        # 保存当前计划到实例变量
        self._current_plan = task_plan

        return Response(
            chat_message=plan_message,
            inner_messages=inner_messages,
        )

    async def _handle_user_feedback(
        self,
        last_message: BaseChatMessage,
        agent_name: str,
        model_context,
        model_client,
        model_client_stream: bool,
        cancellation_token: CancellationToken,
        inner_messages: List,
    ) -> AsyncGenerator[BaseAgentEvent | BaseChatMessage | Response, None]:
        """
        处理用户对计划的反馈
        """
        try:
            # 解析用户反馈
            feedback = self._task_planner.parse_user_feedback(last_message)

            if not feedback.get("accepted"):
                # 用户拒绝计划或提供了修改意见
                if feedback.get("feedback"):
                    # 根据反馈重新规划
                    revised_plan = await self._task_planner.replan_with_feedback(
                        original_plan=self._current_plan,
                        user_feedback=feedback["feedback"],
                        model_context=model_context,
                        cancellation_token=cancellation_token,
                    )

                    # 更新计划
                    self._current_plan = revised_plan
                    self._memory_manager.set_task_plan(revised_plan.model_dump())

                    # 发送新计划
                    plan_message = await self._task_planner.send_plan_to_user(
                        task_plan=revised_plan,
                        agent_name=agent_name,
                    )

                    yield Response(
                        chat_message=plan_message,
                        inner_messages=inner_messages,
                    )
                else:
                    # 用户直接拒绝,没有提供反馈
                    yield Response(
                        chat_message=TextMessage(
                            content="Understood. Please let me know how you'd like to proceed.",
                            source=agent_name,
                            metadata={"internal": "no", "type": "final_answer"},
                        ),
                        inner_messages=inner_messages,
                    )
                return

            # 用户接受计划 → 开始执行
            # 如果用户修改了计划,使用修改后的
            if feedback.get("plan"):
                modified_plan = feedback["plan"]
                self._current_plan.steps = modified_plan.get("steps", self._current_plan.steps)
                self._memory_manager.set_task_plan(self._current_plan.model_dump())

            # 执行任务
            async for event in self._execute_task_plan(
                agent_name=agent_name,
                model_context=model_context,
                model_client=model_client,
                model_client_stream=model_client_stream,
                cancellation_token=cancellation_token,
                inner_messages=inner_messages,
            ):
                yield event

        except Exception as e:
            logger.exception(f"Error handling user feedback")
            yield Response(
                chat_message=TextMessage(
                    content=f"Error processing feedback: {e}",
                    source=agent_name,
                    metadata={"internal": "no", "type": "final_answer"},
                ),
                inner_messages=inner_messages,
            )

    async def _execute_task_plan(
        self,
        agent_name: str,
        model_context,
        model_client,
        model_client_stream: bool,
        cancellation_token: CancellationToken,
        inner_messages: List,
    ) -> AsyncGenerator[BaseAgentEvent | BaseChatMessage | Response, None]:
        """
        执行任务计划 - 核心执行循环
        """
        plan = self._current_plan
        steps = plan.steps
        plan_length = len(steps)

        logger.info(f"Starting task execution with {plan_length} steps")

        # 初始化TodoManager
        todo_items = [
            {
                "content": step.get("title", f"Step {i+1}"),
                "status": "pending",
            }
            for i, step in enumerate(steps)
        ]

        # 更新todo list
        todo_list = self._todo_manager.update(todo_items)
        yield ModelClientStreamingChunkEvent(
            content=f"Task Plan:\n{todo_list}\n\n",
            source=agent_name,
        )

        # 按步骤执行
        for step_index, step in enumerate(steps):
            if self.is_paused:
                raise asyncio.CancelledError()

            # 更新todo状态为in_progress
            todo_items[step_index]["status"] = "in_progress"
            self._todo_manager.update(todo_items)

            # 发送步骤执行状态
            status_message = await self._task_planner.send_step_execution_status(
                step_index=step_index,
                step=step,
                plan_length=plan_length,
                agent_name=agent_name,
                progress_summary=f"Executing: {step.get('title')}",
            )
            yield status_message

            # 执行步骤
            try:
                async for event in self._execute_single_step(
                    step_index=step_index,
                    step=step,
                    agent_name=agent_name,
                    model_context=model_context,
                    model_client=model_client,
                    model_client_stream=model_client_stream,
                    cancellation_token=cancellation_token,
                ):
                    yield event

                # 步骤完成
                todo_items[step_index]["status"] = "completed"
                self._todo_manager.update(todo_items)

            except Exception as e:
                logger.error(f"Error in step {step_index}: {e}")

                # 记录错误
                self._memory_manager.add_error_record({
                    "type": "StepExecutionError",
                    "message": str(e),
                    "step": step_index,
                    "solution": "Attempted auto-recovery or user intervention",
                })

                # 尝试请求用户帮助
                yield ModelClientStreamingChunkEvent(
                    content=f"\n⚠️ Error in step {step_index + 1}: {e}\n",
                    source=agent_name,
                )
                yield ModelClientStreamingChunkEvent(
                    content="Please provide guidance on how to proceed, or I can try an alternative approach.\n\n",
                    source=agent_name,
                )

                # 标记步骤为失败但继续(可以改为等待用户输入)
                todo_items[step_index]["status"] = "completed"
                self._todo_manager.update(todo_items)

        # 所有步骤完成
        self._memory_manager.set_final_result("All tasks completed successfully")
        self._memory_manager.save_current_session()

        final_message = TextMessage(
            content="✅ All tasks have been completed!\n\n" + self._memory_manager.get_session_summary(),
            source=agent_name,
            metadata={"internal": "no", "type": "final_answer"},
        )

        yield Response(
            chat_message=final_message,
            inner_messages=inner_messages,
        )

    async def _execute_single_step(
        self,
        step_index: int,
        step: Dict[str, Any],
        agent_name: str,
        model_context,
        model_client,
        model_client_stream: bool,
        cancellation_token: CancellationToken,
    ) -> AsyncGenerator[BaseAgentEvent | BaseChatMessage, None]:
        """
        执行单个任务步骤

        这里需要智能地决定:
        1. 是否需要加载skill
        2. 是否需要调用tool
        3. 是否需要spawn子智能体
        """
        step_title = step.get("title", f"Step {step_index + 1}")
        step_details = step.get("details", "")
        tool_or_skill = step.get("tool_or_skill")

        yield ModelClientStreamingChunkEvent(
            content=f"### {step_title}\n{step_details}\n\n",
            source=agent_name,
        )

        # 构建步骤执行的prompt
        step_prompt = f"""Execute the following task step:

**Step {step_index + 1}: {step_title}**

Details: {step_details}

{"Suggested tool/skill: " + tool_or_skill if tool_or_skill else ""}

Please execute this step using available tools, skills, or spawn a subagent if needed.
Report the results clearly.
"""

        # 将步骤作为用户消息添加到上下文
        await model_context.add_message(
            UserMessage(content=step_prompt, source="system")
        )

        # 调用LLM执行步骤(这里会自动tool calling)
        turn_count = 0
        max_turns = 5  # 每个步骤最多5轮

        while turn_count < max_turns:
            turn_count += 1

            model_result = None
            async for inference_output in self._call_llm(
                model_client=model_client,
                model_client_stream=model_client_stream,
                system_messages=self._system_messages,
                model_context=model_context,
                workbench=self._workbench,
                handoff_tools=self._handoff_tools,
                manager_tools=self._agent_skills_tools + self._todo_tools + self._subagent_tools,
                agent_name=agent_name,
                cancellation_token=cancellation_token,
                output_content_type=self._output_content_type,
            ):
                if isinstance(inference_output, CreateResult):
                    model_result = inference_output
                else:
                    yield inference_output

            assert model_result is not None

            # 处理thought
            if model_result.thought:
                yield ThoughtEvent(content=model_result.thought, source=agent_name)

            # 添加assistant消息
            await model_context.add_message(
                AssistantMessage(
                    content=model_result.content,
                    source=agent_name,
                    thought=getattr(model_result, "thought", None),
                )
            )

            # 如果是文本响应,步骤完成
            if isinstance(model_result.content, str):
                # 记录步骤执行结果
                self._memory_manager.add_execution_step(
                    step_index=step_index,
                    step_title=step_title,
                    tool_or_skill_used=tool_or_skill,
                    action_details=step_details,
                    result=model_result.content[:500],  # 截取前500字符
                )
                break

            # 否则是tool calls - 处理工具调用
            tool_call_msg = ToolCallRequestEvent(
                content=model_result.content,
                source=agent_name,
                models_usage=model_result.usage,
            )
            yield tool_call_msg

            # 执行工具调用
            async for message in self._handle_tool_calls(
                tool_calls=model_result.content,
                agent_name=agent_name,
                model_context=model_context,
                model_client=model_client,
                model_client_stream=model_client_stream,
                cancellation_token=cancellation_token,
            ):
                yield message

    async def _handle_tool_calls(
        self,
        tool_calls: List,
        agent_name: str,
        model_context,
        model_client,
        model_client_stream: bool,
        cancellation_token: CancellationToken,
    ) -> AsyncGenerator[BaseAgentEvent | BaseChatMessage, None]:
        """
        处理工具调用 - 支持Tool, TodoWrite, Skill, Task(subagent)
        """
        for tool_call in tool_calls:
            argument = json.loads(tool_call.arguments)
            tool_name = tool_call.name

            # 记录工具使用
            self._memory_manager.record_tool_usage(tool_name)

            if tool_name == "TodoWrite":
                async for msg in self.handle_todo_write(
                    argument=argument,
                    agent_name=agent_name,
                    model_context=model_context,
                ):
                    yield msg

            elif tool_name == "Skill":
                # 加载skill
                skill_name = argument.get("skill")
                self._memory_manager.record_skill_loaded(skill_name)

                skill_content = self._skills_loader.run_skill(skill_name)
                await model_context.add_message(
                    UserMessage(
                        content=f"Skill '{skill_name}' loaded:\n{skill_content}",
                        source="system",
                    )
                )
                yield ModelClientStreamingChunkEvent(
                    content=f"📚 Loaded skill: {skill_name}\n",
                    source=agent_name,
                )

            elif tool_name == "Task":
                # 调用子智能体
                agent_type = argument.get("agent_type")
                self._memory_manager.record_subagent_spawned(agent_type)

                async for msg in self.handle_subagent_repsonse(
                    agent_name=agent_name,
                    model_client=model_client,
                    model_client_stream=model_client_stream,
                    model_context=model_context,
                    argument=argument,
                    cancellation_token=cancellation_token,
                    output_content_type=self._output_content_type,
                ):
                    yield msg

            elif tool_name in self._basic_funcs_names:
                # 调用基础工具(run_bash, run_read, etc.)
                async for msg in self._process_model_result(
                    model_result=CreateResult(content=[tool_call], usage=None),
                    inner_messages=[],
                    cancellation_token=cancellation_token,
                    agent_name=agent_name,
                    system_messages=self._system_messages,
                    model_context=model_context,
                    workbench=self._workbench,
                    handoff_tools=self._handoff_tools,
                    handoffs=self._handoffs,
                    model_client=model_client,
                    model_client_stream=model_client_stream,
                    reflect_on_tool_use=self._reflect_on_tool_use,
                    tool_call_summary_format=self._tool_call_summary_format,
                    tool_call_summary_prompt=self._tool_call_summary_prompt,
                    output_content_type=self._output_content_type,
                    format_string=self._output_content_type_format,
                ):
                    yield msg

            else:
                # 未知工具
                await model_context.add_message(
                    UserMessage(
                        content=f"Unknown tool: {tool_name}",
                        source="system",
                    )
                )

    def _extract_keywords_from_request(self, user_request: str) -> List[str]:
        """从用户请求中提取关键词(简单实现)"""
        # 这里可以使用更复杂的NLP方法
        # 简单实现:提取长度>3的单词
        words = user_request.split()
        keywords = [w for w in words if len(w) > 3][:5]  # 最多5个关键词
        return keywords


    async def handle_str_reponse(
            self,
            model_result: CreateResult,
            agent_name: str,
            format_string: str | None,
            inner_messages: List[BaseAgentEvent | BaseChatMessage],
            output_content_type: type[BaseModel] | None,
    ) -> Response:

        if output_content_type:
            content = output_content_type.model_validate_json(model_result.content)
            return Response(
                chat_message=StructuredMessage[output_content_type](  # type: ignore[valid-type]
                    content=content,
                    source=agent_name,
                    models_usage=model_result.usage,
                    format_string=format_string,
                ),
                inner_messages=inner_messages,
            )
        else:
            return Response(
                chat_message=TextMessage(
                    content=model_result.content,
                    source=agent_name,
                    models_usage=model_result.usage,
                ),
                inner_messages=inner_messages,
            )
    
    def get_tools_for_agent(self, agent_type: str) -> list:
        """Filter tools based on agent type."""
        allowed = self._sub_agent_config.get(agent_type, {}).get("tools", "*")
        if allowed == "*":
            return self._tools
        return [t for t in self._tools if t.name in allowed]

    # TODO: handle Task
    async def handle_subagent_repsonse(
        self,
        agent_name: str,
        model_client: ChatCompletionClient,
        model_client_stream: bool,
        model_context: ChatCompletionContext,
        argument:Dict[str, Any],
        cancellation_token: CancellationToken,
        output_content_type: type[BaseModel] | None,
    ) -> AsyncGenerator[BaseAgentEvent | BaseChatMessage, None]:
        """
        Sub agent can actuate the sub task.

        The types of sub agent:
        1. code_executor
        2. normal drsai agent
        """
        try:
            description, prompt, agent_type = argument["description"], argument["prompt"], argument["agent_type"]

            # get sub agent system prompt
            sub_system = f"""You are a {agent_type} subagent at {self._work_dir}.

    {self._sub_agent_config[agent_type].get("prompt", "")}

    Complete the task and return a clear, concise summary."""
            
            # construct task messages
            task_messages: Sequence[BaseChatMessage] = []
            task_messages.append(TextMessage(content=f"Current task: \n\n{prompt}", source="user"))
            llm_messages = await model_context.get_messages()
            backgroud_message = "Below are the historical chat records between the user and various intelligent assistants, which can be referenced when executing the current task.\n\n"
            for llm_message in llm_messages:
                if isinstance(llm_message, UserMessage) or isinstance(llm_message, AssistantMessage):
                    backgroud_message += f"{llm_message.source}: {llm_message.content}\n\n"
            task_messages.append(TextMessage(content=backgroud_message, source="user"))

            #  TODO: handle turn count fro multi-turn task.
            # turn_count = 0
            # while turn_count < self._max_turn_count:
            #     turn_count += 1
        
            if agent_type == "coder_executor":
                subagent = CodeExecutorAgent(
                    name=agent_type,
                    code_executor=self._executor,
                    model_client_stream=model_client_stream,
                )
            else:
                tools = self.get_tools_for_agent(agent_type)
                description=self._sub_agent_config[agent_type].get("description", "")
                subagent = DrSaiAgent(
                    name=agent_type,
                    system_message=sub_system,
                    description=description,
                    tools=tools,
                    model_client=model_client,
                    model_client_stream=model_client_stream,
                    output_content_type=output_content_type,)
            
            async for message in subagent.on_messages_stream(messages=task_messages, cancellation_token=cancellation_token):
                if isinstance(message, Response):
                    yield message.chat_message
                    await model_context.add_message(
                        UserMessage(
                            content=message.chat_message.content,
                            source="user",
                        )
                    )
                    return
                yield message

        except Exception as e:
            logger.exception(f"Error in {self.name}")
            yield ModelClientStreamingChunkEvent(
                content=str(e)+"\n\n",
                source=self.name,
            )
            await model_context.add_message(
                UserMessage(
                    content=str(e),
                    source="user",
                )
            )
            yield StopMessage(
                content=str(e),
                source=agent_name,
            )

    # TODO: handle TodoWrite
    async def handle_todo_write(
        self,
        argument: Dict[str, Any],
        agent_name: str,
        model_context: ChatCompletionContext,
    ) -> AsyncGenerator[BaseAgentEvent | BaseChatMessage, None]:
        try:
            todo_list = self._todo_manager.update(argument["items"])
            # send stream message
            yield ModelClientStreamingChunkEvent(
                content=todo_list+"\n\n",
                source=self.name,
            )
            # add message to model_context with user source
            await model_context.add_message(
                UserMessage(
                    content=self._todo_manager.get_task_prompt(),
                    source="user",
                )
            )
            # send text message to save to db in drsai ui
            yield TextMessage(
                content=todo_list,
                source=agent_name,
            )
        except Exception as e:
            logger.exception(f"Error in {self.name}")
            yield ModelClientStreamingChunkEvent(
                content=str(e)+"\n\n",
                source=self.name,
            )
            await model_context.add_message(
                UserMessage(
                    content=str(e),
                    source="user",
                )
            )
            yield StopMessage(
                content=str(e),
                source=agent_name,
            )

    def _to_config(self) -> EdgeAgentConfig:
        """Convert the assistant agent to a declarative config."""

        return EdgeAgentConfig(
            name=self.name,
            model_client=self._model_client.dump_component(),
            tools=None,  # versionchanged:: v0.5.5  Now tools are not serialized, Cause they are part of the workbench.
            workbench=self._workbench.dump_component() if self._workbench else None,
            handoffs=list(self._handoffs.values()) if self._handoffs else None,
            model_context=self._model_context.dump_component(),
            memory=[memory.dump_component() for memory in self._memory] if self._memory else None,
            description=self.description,
            system_message=self._system_messages[0].content
            if self._system_messages and isinstance(self._system_messages[0].content, str)
            else None,
            model_client_stream=self._model_client_stream,
            reflect_on_tool_use=self._reflect_on_tool_use,
            tool_call_summary_format=self._tool_call_summary_format,
            tool_call_summary_prompt=self._tool_call_summary_prompt,
            structured_message_factory=self._structured_message_factory.dump_component()
            if self._structured_message_factory
            else None,
            metadata=self._metadata,
            db_manager_config=self._db_manager.dump_component(),
            thread_id=self._thread_id,
            user_id=self._user_id,
            skills_dir=self._skills_dir,
        )
    
    @classmethod
    def _from_config(
        cls, config: EdgeAgentConfig, 
        db_manager: DatabaseManager,
        memory_function: Callable = None,
        reply_function: Callable = None,
        **kwargs,
        ) -> Self:
        """Create an assistant agent from a declarative config."""
        if config.structured_message_factory:
            structured_message_factory = StructuredMessageFactory.load_component(config.structured_message_factory)
            format_string = structured_message_factory.format_string
            output_content_type = structured_message_factory.ContentModel

        else:
            format_string = None
            output_content_type = None

        return cls(
            name=config.name,
            model_client=ChatCompletionClient.load_component(config.model_client),
            workbench=Workbench.load_component(config.workbench) if config.workbench else None,
            handoffs=config.handoffs,
            model_context=ChatCompletionContext.load_component(config.model_context) if config.model_context else None,
            tools=[BaseTool.load_component(tool) for tool in config.tools] if config.tools else None,
            memory=[Memory.load_component(memory) for memory in config.memory] if config.memory else None,
            description=config.description,
            system_message=config.system_message,
            model_client_stream=config.model_client_stream,
            reflect_on_tool_use=config.reflect_on_tool_use,
            tool_call_summary_format=config.tool_call_summary_format,
            tool_call_summary_prompt=config.tool_call_summary_prompt,
            output_content_type=output_content_type,
            output_content_type_format=format_string,
            metadata=config.metadata,
            memory_function=memory_function,
            reply_function=reply_function,
            db_manager=db_manager,
            thread_id=config.thread_id,
            user_id=config.user_id,
            skills_dir=config.skills_dir,
            **kwargs,
        )



"""
EdgeAgent Long-term Memory Manager
长期记忆管理模块

整合:
1. UserProfileManager - 管理本地文件存储的记忆
2. DrSaiChatCompletionContext - 管理当前会话的上下文压缩
3. 提供记忆检索和学习功能
"""

from typing import List, Optional, Dict, Any
from pathlib import Path
from loguru import logger

from drsai.modules.components.model_context import ChatCompletionContext
from drsai.modules.baseagent import UserMessage
from .user_profile_manager import (
    UserProfileManager,
    SessionMemory,
    TaskStep,
)


class LongTermMemoryManager:
    """
    长期记忆管理器

    职责:
    1. 管理当前会话的SessionMemory对象
    2. 在会话结束时保存记忆到本地文件
    3. 在需要时检索历史记忆辅助任务规划
    4. 支持从成功任务中学习并保存为skills
    """

    def __init__(
        self,
        user_profile_manager: UserProfileManager,
        model_context: ChatCompletionContext,
        session_id: str,
    ):
        """
        Args:
            user_profile_manager: 用户配置管理器
            model_context: 模型上下文(用于管理当前会话)
            session_id: 当前会话ID (thread_id)
        """
        self.user_profile_manager = user_profile_manager
        self.model_context = model_context
        self.session_id = session_id

        # 当前会话的记忆对象
        self.current_session_memory: Optional[SessionMemory] = None

        # 尝试加载已存在的会话记忆
        self._load_or_create_session_memory()

    def _load_or_create_session_memory(self):
        """加载或创建当前会话的记忆对象"""
        # 尝试加载
        existing_memory = self.user_profile_manager.load_session_memory(self.session_id)

        if existing_memory:
            self.current_session_memory = existing_memory
            logger.info(f"Loaded existing session memory for {self.session_id}")
        else:
            # 创建新的(等待用户请求后初始化)
            self.current_session_memory = None
            logger.info(f"Will create new session memory for {self.session_id}")

    def start_session(self, user_request: str, needs_plan: bool = False):
        """
        开始新会话或更新会话信息

        Args:
            user_request: 用户请求
            needs_plan: 是否需要任务规划
        """
        if self.current_session_memory is None:
            self.current_session_memory = self.user_profile_manager.create_session_memory(
                session_id=self.session_id,
                user_request=user_request,
                needs_plan=needs_plan,
            )
            logger.info(f"Started new session memory for {self.session_id}")
        else:
            # 更新已有会话的信息
            if not self.current_session_memory.original_user_request:
                self.current_session_memory.original_user_request = user_request
            self.current_session_memory.needs_plan = needs_plan

    def set_task_plan(self, task_plan: Dict[str, Any]):
        """
        设置任务规划

        Args:
            task_plan: 任务规划字典
        """
        if self.current_session_memory:
            self.current_session_memory.task_plan = task_plan
            logger.info(f"Set task plan with {len(task_plan.get('steps', []))} steps")

    def add_execution_step(
        self,
        step_index: int,
        step_title: str,
        tool_or_skill_used: Optional[str],
        action_details: str,
        result: str,
        error: Optional[str] = None,
    ):
        """
        添加执行步骤记录

        Args:
            step_index: 步骤索引
            step_title: 步骤标题
            tool_or_skill_used: 使用的工具或技能
            action_details: 操作详情
            result: 执行结果
            error: 错误信息(如果有)
        """
        if not self.current_session_memory:
            logger.warning("Cannot add execution step: session not started")
            return

        step = TaskStep(
            step_index=step_index,
            step_title=step_title,
            tool_or_skill_used=tool_or_skill_used,
            action_details=action_details,
            result=result,
            error=error,
        )

        self.current_session_memory.execution_steps.append(step)

        # 更新工具/技能使用记录
        if tool_or_skill_used:
            if tool_or_skill_used not in self.current_session_memory.tools_used:
                self.current_session_memory.tools_used.append(tool_or_skill_used)

    def record_tool_usage(self, tool_name: str):
        """记录工具使用"""
        if self.current_session_memory and tool_name not in self.current_session_memory.tools_used:
            self.current_session_memory.tools_used.append(tool_name)

    def record_skill_loaded(self, skill_name: str):
        """记录加载的skill"""
        if self.current_session_memory and skill_name not in self.current_session_memory.skills_loaded:
            self.current_session_memory.skills_loaded.append(skill_name)

    def record_subagent_spawned(self, agent_type: str):
        """记录调用的子智能体"""
        if self.current_session_memory and agent_type not in self.current_session_memory.subagents_spawned:
            self.current_session_memory.subagents_spawned.append(agent_type)

    def add_error_record(self, error_info: Dict[str, Any]):
        """
        记录错误及修正过程

        Args:
            error_info: 错误信息字典,应包含 {type, message, step, solution}
        """
        if self.current_session_memory:
            self.current_session_memory.errors_encountered.append(error_info)

    def set_final_result(self, result: str):
        """
        设置最终结果

        Args:
            result: 最终结果描述
        """
        if self.current_session_memory:
            from datetime import datetime
            self.current_session_memory.final_result = result
            self.current_session_memory.end_time = datetime.now().isoformat()

    def add_learned_pattern(self, pattern: str):
        """
        添加学到的模式/经验

        Args:
            pattern: 学到的模式描述
        """
        if self.current_session_memory:
            self.current_session_memory.learned_patterns.append(pattern)

    def save_current_session(self) -> str:
        """
        保存当前会话记忆到文件

        Returns:
            保存的文件路径
        """
        if not self.current_session_memory:
            logger.warning("No session memory to save")
            return ""

        filepath = self.user_profile_manager.save_session_memory(self.current_session_memory)
        logger.info(f"Saved current session memory to {filepath}")
        return filepath

    async def retrieve_relevant_memories(
        self,
        keywords: List[str],
        limit: int = 3,
    ) -> List[SessionMemory]:
        """
        检索相关的历史记忆

        Args:
            keywords: 搜索关键词
            limit: 返回数量限制

        Returns:
            相关的SessionMemory列表
        """
        memories = self.user_profile_manager.search_memories(keywords=keywords, limit=limit)
        logger.info(f"Retrieved {len(memories)} relevant memories for keywords: {keywords}")
        return memories

    async def inject_relevant_memories_to_context(
        self,
        keywords: List[str],
        limit: int = 2,
    ):
        """
        检索相关记忆并注入到当前上下文中

        Args:
            keywords: 搜索关键词
            limit: 返回数量限制
        """
        memories = await self.retrieve_relevant_memories(keywords, limit)

        if memories:
            memory_text = "=== Relevant Historical Memories ===\n\n"
            for i, mem in enumerate(memories, 1):
                memory_text += f"**Memory {i}** (Session: {mem.session_id}):\n"
                memory_text += f"- Task: {mem.original_user_request}\n"
                memory_text += f"- Approach: {mem.plan_summary if mem.task_plan else 'Direct execution'}\n"
                memory_text += f"- Tools Used: {', '.join(mem.tools_used)}\n"
                if mem.learned_patterns:
                    memory_text += f"- Key Learnings: {'; '.join(mem.learned_patterns)}\n"
                memory_text += "\n"

            # 将记忆作为UserMessage添加到上下文
            await self.model_context.add_message(
                UserMessage(
                    content=memory_text,
                    source="memory_system"
                )
            )
            logger.info(f"Injected {len(memories)} memories into context")

    async def learn_from_session_and_save_skill(
        self,
        skill_name: str,
        skill_description: str,
    ) -> str:
        """
        从当前会话学习并保存为skill

        Args:
            skill_name: skill名称
            skill_description: skill描述

        Returns:
            保存的skill文件路径
        """
        if not self.current_session_memory:
            logger.warning("No session memory to learn from")
            return ""

        # 构建SKILL.md内容
        skill_content = f"""---
name: {skill_name}
description: {skill_description}
---

# {skill_name}

## Task Context
This skill was learned from session: {self.current_session_memory.session_id}

Original User Request:
{self.current_session_memory.original_user_request}

## Approach

"""
        # 添加任务规划
        if self.current_session_memory.task_plan:
            skill_content += "### Task Breakdown\n"
            steps = self.current_session_memory.task_plan.get('steps', [])
            for i, step in enumerate(steps, 1):
                skill_content += f"{i}. **{step.get('title', 'Step')}**: {step.get('details', '')}\n"
            skill_content += "\n"

        # 添加工具和技能使用
        if self.current_session_memory.tools_used:
            skill_content += "### Tools & Skills Used\n"
            for tool in self.current_session_memory.tools_used:
                skill_content += f"- {tool}\n"
            skill_content += "\n"

        # 添加关键步骤
        if self.current_session_memory.execution_steps:
            skill_content += "### Key Execution Steps\n"
            for step in self.current_session_memory.execution_steps:
                skill_content += f"**Step {step.step_index + 1}: {step.step_title}**\n"
                skill_content += f"- Tool/Skill: {step.tool_or_skill_used or 'N/A'}\n"
                skill_content += f"- Action: {step.action_details}\n"
                skill_content += f"- Result: {step.result}\n"
                if step.error:
                    skill_content += f"- Error Handled: {step.error}\n"
                skill_content += "\n"

        # 添加学到的经验
        if self.current_session_memory.learned_patterns:
            skill_content += "### Key Learnings\n"
            for pattern in self.current_session_memory.learned_patterns:
                skill_content += f"- {pattern}\n"
            skill_content += "\n"

        # 添加错误处理经验
        if self.current_session_memory.errors_encountered:
            skill_content += "### Error Handling\n"
            for error in self.current_session_memory.errors_encountered:
                skill_content += f"- **{error.get('type', 'Error')}**: {error.get('message', '')}\n"
                if error.get('solution'):
                    skill_content += f"  Solution: {error['solution']}\n"
            skill_content += "\n"

        # 保存skill
        filepath = self.user_profile_manager.save_learned_skill(skill_name, skill_content)
        logger.info(f"Learned and saved skill '{skill_name}' from session {self.session_id}")

        # 记录到当前会话
        self.add_learned_pattern(f"Created skill: {skill_name}")

        return filepath

    def get_session_summary(self) -> str:
        """
        获取当前会话的摘要

        Returns:
            会话摘要文本
        """
        if not self.current_session_memory:
            return "No active session"

        summary = f"""### Session Summary ({self.session_id})

**User Request:** {self.current_session_memory.original_user_request}

**Planning:** {'Yes' if self.current_session_memory.needs_plan else 'No'}

**Steps Executed:** {len(self.current_session_memory.execution_steps)}

**Tools Used:** {', '.join(self.current_session_memory.tools_used) if self.current_session_memory.tools_used else 'None'}

**Skills Loaded:** {', '.join(self.current_session_memory.skills_loaded) if self.current_session_memory.skills_loaded else 'None'}

**Errors:** {len(self.current_session_memory.errors_encountered)}

**Status:** {'Completed' if self.current_session_memory.end_time else 'In Progress'}
"""
        return summary


if __name__ == "__main__":
    # 测试代码
    from drsai.modules.components.model_context import BufferedChatCompletionContext

    # 创建测试环境
    user_manager = UserProfileManager("/tmp/test_edgeagent", "test_user_001")
    context = BufferedChatCompletionContext(buffer_size=10)

    memory_manager = LongTermMemoryManager(
        user_profile_manager=user_manager,
        model_context=context,
        session_id="test_session_001",
    )

    # 模拟会话
    memory_manager.start_session(
        user_request="测试数据分析任务",
        needs_plan=True
    )

    memory_manager.set_task_plan({
        "needs_plan": True,
        "steps": [
            {"title": "读取数据", "details": "读取CSV文件"},
            {"title": "数据分析", "details": "计算统计信息"},
        ]
    })

    memory_manager.add_execution_step(
        step_index=0,
        step_title="读取数据",
        tool_or_skill_used="run_read",
        action_details="读取文件test.csv",
        result="成功读取1000行数据"
    )

    memory_manager.set_final_result("任务完成")
    memory_manager.save_current_session()

    print("Memory manager test completed")
    print(memory_manager.get_session_summary())



from pathlib import Path
import subprocess
import re

# Dangerous command patterns (regex)
_DANGEROUS_PATTERNS = [
    r'\bsudo\b',
    r'\bsu\s',
    r'\bshutdown\b',
    r'\breboot\b',
    r'\bhalt\b',
    r'\bmkfs\b',
    r'\bmknod\b',
    r'\bdd\b.+\bof=/dev\b',      # dd writing to block devices
    r'\bchmod\b.+\b[0-7]*[67]{1}[0-7]{2}\s+/',  # chmod with write perms on root paths
    r'\bchown\b.+/',
    r'\bcrontab\b',
    r'\bkillall\b',
    r'\biptables\b',
    r'>\s*/etc/',                 # redirect to /etc
    r'>\s*/dev/',                 # redirect to devices
    r'>\s*/sys/',
    r'>\s*/proc/',
    # --- rm variants ---
    r'\brm\s+(-[a-zA-Z]*r[a-zA-Z]*f|-[a-zA-Z]*f[a-zA-Z]*r)\s+/',  # rm -rf /...
    r'\brm\s+-[a-zA-Z]*r[a-zA-Z]*\s+/',                             # rm -r /...
    r'\brm\s+-[a-zA-Z]*f[a-zA-Z]*\s+/',                             # rm -f /... (force, no recursion)
    # --- find with deletion actions ---
    r'\bfind\b.+\-delete\b',                    # find -delete
    r'\bfind\b.+\-exec\s+rm\b',                 # find -exec rm ...
    r'\bfind\b.+\-exec\s+unlink\b',             # find -exec unlink ...
    # --- xargs piped to rm/unlink ---
    r'\bxargs\b.*\brm\b',                        # xargs rm (e.g. find ... | xargs rm)
    r'\bxargs\b.*\bunlink\b',                    # xargs unlink
    # --- low-level / secure deletion tools ---
    r'\bunlink\b',                               # unlink (syscall wrapper, deletes file)
    r'\bshred\b',                                # shred (overwrite + delete)
    r'\bwipe\b',                                 # wipe (secure delete)
    r'\bsrm\b',                                  # srm (secure-delete package)
    # --- directory removal ---
    r'\brmdir\b',                                # rmdir (remove empty dirs)
    # --- file truncation to zero ---
    r'\btruncate\b.*(-s\s*0|--size[=\s]*0)\b',  # truncate --size 0 / -s 0
    # --- move to /dev/null (effectively destroys content) ---
    r'\bmv\b.+/dev/null\b',
]
_DANGEROUS_RE = re.compile('|'.join(_DANGEROUS_PATTERNS), re.IGNORECASE)

# Regex to extract absolute paths from shell commands
_ABS_PATH_RE = re.compile(r'(?:^|[\s=\'",;|&<>(){}])(/(?:[^\s;|&><\'"\\{}()]+))')


def get_operator_funcs(worker_dir: str|Path, extra_dirs: list[str|Path] = None, only_in_workspace: bool = True )->list[callable]:

    WORKDIR = Path(worker_dir).resolve()
    ALLOWED_DIRS = [WORKDIR] + [Path(d).resolve() for d in (extra_dirs or [])]

    def safe_path(p: str) -> Path:
        """Ensure path stays within workspace or allowed directories."""
        resolved = Path(p).resolve()
        # If path is absolute, check directly against allowed dirs
        if Path(p).is_absolute():
            if only_in_workspace and not any(resolved.is_relative_to(d) for d in ALLOWED_DIRS):
                raise ValueError(f"Path escapes workspace: {p}")
            return resolved
        # Relative path: resolve against WORKDIR
        path = (WORKDIR / p).resolve()
        if only_in_workspace and not any(path.is_relative_to(d) for d in ALLOWED_DIRS):
            raise ValueError(f"Path escapes workspace: {p}")
        return path

    def _check_cmd_paths(cmd: str) -> str | None:
        """Return an error string if any absolute path in cmd escapes allowed dirs, else None."""
        for match in _ABS_PATH_RE.finditer(cmd):
            raw = match.group(1).rstrip('/')
            if not raw:
                continue
            try:
                resolved = Path(raw).resolve()
            except Exception:
                continue
            if not any(resolved.is_relative_to(d) for d in ALLOWED_DIRS):
                return f"Error: Path '{raw}' is outside the allowed workspace"
        return None

    # Mutable current directory state (persists across run_bash calls)
    _cwd = [WORKDIR]

    def run_bash(cmd: str) -> str:
        """Execute shell command in workspace directory.

        The working directory persists across calls: cd commands take effect
        for subsequent invocations, as long as the target stays within the
        allowed workspace.
        """
        # Check dangerous patterns
        if _DANGEROUS_RE.search(cmd):
            return "Error: Dangerous command detected"
        # Check absolute paths referenced in command
        if only_in_workspace:
            path_err = _check_cmd_paths(cmd)
            if path_err:
                return path_err
        try:
            # Append a sentinel so we can capture the resulting directory
            wrapped = f'{cmd}\necho "__DRSAI_CWD__:$(pwd)"'
            r = subprocess.run(
                wrapped, shell=True, cwd=_cwd[0],
                capture_output=True, text=True, timeout=90
            )
            raw = r.stdout + r.stderr
            # Parse and strip the sentinel line
            lines = raw.splitlines()
            out_lines = []
            for line in lines:
                if line.startswith("__DRSAI_CWD__:"):
                    new_dir = Path(line[len("__DRSAI_CWD__:"):]).resolve()
                    # Only update cwd if still within allowed dirs
                    if any(new_dir.is_relative_to(d) for d in ALLOWED_DIRS):
                        _cwd[0] = new_dir
                    elif only_in_workspace:
                        out_lines.append(
                            f"Warning: cd target '{new_dir}' is outside workspace; "
                            "cwd not updated"
                        )
                else:
                    out_lines.append(line)
            return ("\n".join(out_lines).strip() or "(no output)")[:50000]
        except Exception as e:
            return f"Error: {e}"


    def run_read(path: str, minilimit: int = None, maxlimit: int = -1) -> str:
        """
        Read file contents.
        
        Args:
            path : Path to file.
            minilimit : The start of  Maximum number of lines to read.
            maxlimit : The end of  Maximum number of lines to read.
        """
        try:
            lines = safe_path(path).read_text().splitlines()
            if minilimit:
                lines = lines[minilimit:maxlimit]
            return "\n".join(lines)[:50000]
        except Exception as e:
            return f"Error: {e}"


    def run_write(path: str, content: str) -> str:
        """Write content to file."""
        try:
            fp = safe_path(path)
            fp.parent.mkdir(parents=True, exist_ok=True)
            fp.write_text(content)
            return f"Wrote {len(content)} bytes to {path}"
        except Exception as e:
            return f"Error: {e}"


    def run_edit(path: str, old_text: str, new_text: str) -> str:
        """Replace exact text in file."""
        try:
            fp = safe_path(path)
            text = fp.read_text()
            if old_text not in text:
                return f"Error: Text not found in {path}"
            fp.write_text(text.replace(old_text, new_text, 1))
            return f"Edited {path}"
        except Exception as e:
            return f"Error: {e}"
    
    return [run_bash, run_read, run_write, run_edit]


from drsai import DrSai
import os, sys
from pathlib import Path
here = Path(__file__).parent
from typing import Generator, Optional, Union, Dict, Any
from fastapi import FastAPI, Request, Header
from fastapi.responses import StreamingResponse # , HTMLResponse, FileResponse, JSONResponse, PlainTextResponse
from fastapi import FastAPI, HTTPException #, Form, File, UploadFile, Response
from fastapi import APIRouter # , Query
import traceback
import inspect  
from collections.abc import AsyncGenerator 

try:
    from drsai.version import __version__
except:
    sys.path.append(str(here.parent.parent))
    from drsai.version import __version__

from loguru import logger
logger = logger.bind(name="app_worker.py")

class DrSaiAPP(DrSai):
    '''
    chat/completion:路由,直接处理聊天和自动回复请求。
    该路由接收前端页面的请求，并返回相应的回复。
    该路由的请求参数包括:
    - messages: 输入的消息列表

    OpenAI Assistants格式的标准接口的DrSai后端服务:
    该后端服务通过http请求或者hepai的openai assistants格式的标准接口api, 提供Dr.Sai多智能体后端服务。
    包括:
    1. Assistants-相关接口，包括创建、获取、删除、更新助手。用于对接前后端Agents设置和前端页面的交互。
    2. Threads-相关接口，包括创建、获取、删除、更新会话。用于对接前端页面的会话交互。
    3. Runs-相关接口，包括创建、获取、删除、更新运行。用于对接前端页面的运行交互。

    '''
    app = FastAPI(docs_url="/docs", redoc_url=None)
    router = APIRouter(prefix="/apiv2", tags=["openai_style_api"])
    
    def __init__(self, **kwargs):

        super(DrSaiAPP, self).__init__(**kwargs)

        self._init_router()
        self.app.include_router(DrSaiAPP.router)

        self._info = {}
    
    def _init_router(self):

        # 测试路由
        DrSaiAPP.router.get("/")(self.index)
        DrSaiAPP.router.get("/models")(self.list_models)

        # chat/completion路由
        DrSaiAPP.router.post("/chat/completions")(self.a_chat_completions)
        # TODO: 增加支持pause/resume/close会话的接口

        # agents/groupchat测试路由
        DrSaiAPP.router.get("/agents/get_info")(self.a_get_agents_info)
        DrSaiAPP.router.post("/agents/test_api")(self.a_agents_test_api)
        DrSaiAPP.router.get("/agents/list_agents")(self.a_list_agents)



    #### --- 关于DrSai的路由 --- ####
    async def index(self, request: Request):
        return f"Hello, world! This is DrSai WebUI {__version__}"
    
    async def list_models(self, request: Request):

        agents_info = await self.get_agents_info()
        models_data = []
        for index, agent_info in enumerate(agents_info):
            agent_name = agent_info.get("name", "unknown")
            models_data.append({
                "id": agent_name,
                "object": "agent",
                "created": None,
                "owned_by": None
                })
        models = {
            "object": "list",
            "data": models_data
            }
        return models
    
    async def a_list_agents(self, request: Request):
        models_data = [{
                "id": self._info["name"],
                "object": "agent",
                "owner": self._info["author"],
                "description": self._info["description"],
                "version": self._info["version"],
                }]
        
        models = {
            "object": "list",
            "data": models_data
            }
        return models
    
    ### --- 关于chat_completions的路由 --- ###
    async def a_chat_completions(self, request: Request):

        headers = request.headers
        if not isinstance(headers, dict):
            headers = dict(headers)
        authorization = headers.get("authorization", None)
        if authorization:
            apikey = authorization.split(" ")[-1]
        else:
            apikey = None

        # apikey = headers.get("authorization").split(" ")[-1]
        params = await request.json()
        params.update({"apikey": apikey})
        if "messages" not in params or "model" not in params:
            raise HTTPException(status_code=400, detail="messages and model must be required, see https://platform.openai.com/docs/api-reference/chat")
        # return self.try_except_raise_http_exception(
        #     self.a_start_chat_completions, **params
        #     )
        return await self.try_except_raise_http_exception(
            self.a_start_chat_completions, **params
            )
    
     ### --- 关于agents/groupchat的测试路由 --- ###
    async def a_get_agents_info(self, request: Request) -> list[dict[str, Any]]:
        """
        获取agents/groupchat的相关信息
        """
        headers = request.headers
        if not isinstance(headers, dict):
            headers = dict(headers)
        authorization = headers.get("authorization", None)
        if authorization:
            apikey = authorization.split(" ")[-1]
        else:
            apikey = None
            
        if apikey != self.drsai_test_api_key:
            return {"error": "Invalid drsai_test_api_key"}
        
        return await self.try_except_raise_http_exception(
            self.get_agents_info,
        )

    async def a_agents_test_api(self, request: Request):
        """
        groupchat/agents的测试端口
        """
        headers = request.headers
        if not isinstance(headers, dict):
            headers = dict(headers)
        authorization = headers.get("authorization", None)
        if authorization:
            apikey = authorization.split(" ")[-1]
        else:
            apikey = None

        # apikey = headers.get("authorization").split(" ")[-1]
        params = await request.json()
        params.update({"apikey": apikey})
        if "messages" not in params or "model" not in params:
            raise HTTPException(status_code=400, detail="messages and model must be required, see https://platform.openai.com/docs/api-reference/chat")
        # return self.try_except_raise_http_exception(
        #     self.a_start_chat_completions, **params
        #     )
        return await self.try_except_raise_http_exception(
            self.test_agents, **params
            )

    ### --- 其它函数 --- #### 
    async def try_except_raise_http_exception(self, func, *args, **kwargs):  # 改为异步函数
        """智能捕获函数内部raise的异常，转换为HTTPException返回，支持同步/异步函数"""
        try:
            # 判断是否是异步函数
            if inspect.iscoroutinefunction(func):
                res = await func(*args, **kwargs)  # 异步函数需要await
            else:
                res = func(*args, **kwargs)  # 同步函数直接调用

            # 处理同步/异步生成器
            if isinstance(res, (AsyncGenerator, Generator)):  # 同时检查两种生成器
                return StreamingResponse(res)
            return res
    
        except Exception as e:
            tb_str = traceback.format_exception(*sys.exc_info())
            tb_str = "".join(tb_str)
            logger.debug(f"Error: {e}.\nTraceback: {tb_str}")

            e_class = e.__class__.__name__
            error_mapping = {
                "ModuleNotFoundError": 404,
                "NotImplementedError": 501,
                "ValueError": 400,
                "TypeError": 400,
                "KeyError": 400,
                "AttributeError": 400,
                "HTTPException": 400,
                "AssertionError": 400,
                "PermissionError": 403,
                "FileNotFoundError": 404,
                "OSError": 500,
                "RuntimeError": 500,
                # 添加更多映射...
            }
            status_code = error_mapping.get(e_class, 400)
            raise HTTPException(
                status_code=status_code,
                detail=f'{e_class}("{str(e)}")'
            )



from .user_profile_manager import UserProfileManager
from .todo_manager import TodoManager
from .operater_funs import get_operator_funcs
from .task_planner import TaskPlanner
from .memory_manager import LongTermMemoryManager
from .get_managers_tools import (
    create_local_venv,
    get_agent_skills_tool,
    get_subagent_tools,
    get_todo_manager_tool,
)


from autogen_core import (
    AgentId,
    AgentInstantiationContext,
    AgentRuntime,
    AgentType,
    DefaultTopicId,
    MessageContext,
    event,
    rpc,
    SingleThreadedAgentRuntime,
    TypeSubscription,
    CancellationToken,
    ComponentLoader,
    ComponentFromConfig,
    Component,
    ComponentModel,
    ComponentBase,
    FunctionCall,
    
)



from typing_extensions import Unpack
import os
import logging
import warnings
from typing import (
    Any,
    AsyncGenerator,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Type,
    Union,
    cast,
)
from typing_extensions import Required
from pydantic import BaseModel

from openai.types.chat import ChatCompletionChunk
from tiktoken.model import MODEL_TO_ENCODING

from autogen_core import (
    EVENT_LOGGER_NAME,
    TRACE_LOGGER_NAME,
    CancellationToken,
    Component,
    FunctionCall,
    Image,
)
from autogen_ext.models._utils.normalize_stop_reason import normalize_stop_reason
from autogen_ext.models._utils.parse_r1_content import parse_r1_content
from autogen_core.models import (
    ChatCompletionTokenLogprob,
    CreateResult,
    LLMMessage,
    ModelInfo,
    ModelFamily,
    RequestUsage,
    TopLogprob,
    UserMessage,
    SystemMessage,
    AssistantMessage,
)
from autogen_core.logging import LLMCallEvent, LLMStreamEndEvent, LLMStreamStartEvent
from autogen_core.tools import Tool, ToolSchema

from autogen_ext.models.openai import OpenAIChatCompletionClient
from autogen_ext.models.openai._model_info import (
    _MODEL_TOKEN_LIMITS,
)
from autogen_ext.models.openai._openai_client import (
    convert_tools, 
    _add_usage,
    to_oai_type,
    create_kwargs,
    count_tokens_openai,
    _model_info,
    )
from autogen_ext.models.openai.config import (
    OpenAIClientConfiguration, 
    OpenAIClientConfigurationConfigModel)

logger = logging.getLogger(EVENT_LOGGER_NAME)


class HepAIModelInfo(ModelInfo):
    token_model: Required[str]

# See OpenAI docs for explanation of these parameters
class HepAIClientConfiguration(OpenAIClientConfiguration, total=False):
    api_version: str | None = None


class HepAIClientConfigurationConfigModel(OpenAIClientConfigurationConfigModel):
    api_version: str | None = None


class HepAIChatCompletionClient(OpenAIChatCompletionClient, Component[HepAIClientConfigurationConfigModel]):

    component_type = "model"
    component_config_schema = HepAIClientConfigurationConfigModel
    component_provider_override = "drsai.HepAIChatCompletionClient"

    def __init__(self, **kwargs: Unpack[HepAIClientConfiguration]):

        if "api_key" not in kwargs:
            kwargs["api_key"] = os.environ.get("HEPAI_API_KEY")
        if "base_url" not in kwargs:
            kwargs["base_url"] = "https://aiapi.ihep.ac.cn/apiv2"

        if "model_info" not in kwargs:
            model_info: Optional[HepAIModelInfo] ={
                "vision": False,
                "function_calling": False,  # You must sure that the model can handle function calling
                "json_output": False,
                "structured_output": False,
                "family": ModelFamily.UNKNOWN,
                "multiple_system_messages":False,
                "token_model": "gpt-4o-2024-11-20", # Default model for token counting
            }
            kwargs["model_info"] = model_info
        self.r1_series = [
            "aliyun/qwen-turbo-latest",
            "aliyun/qwen3-30b-a3b",
            "aliyun/qwen3-235b-a22b",
            "aliyun/qwq-plus-latest",
            "anthropic/claude-sonnet-4-thinking",
            "anthropic/claude-opus-4-thinking",
            "deepseek-ai/deepseek-r1:671b",
            "deepseek-ai/deepseek-r1:32b",
            "deepseek-ai/deepseek-r1:7b",

        ]
        self.v3_series = [
            "aliyun/qwen-max-latest",
            "aliyun/qwen-coder-plus-latest",
            "aliyun/qwen-long-latest",
            "openai/o1-mini",
            "openai/computer-use-preview",
            "anthropic/claude-sonnet-4",
            "anthropic/claude-3-5-haiku",
            "anthropic/claude-opus-4",
            "deepseek-ai/deepseek-v3:671b"

        ]

        self.vsion_series = [
            "aliyun/qwen-vl-max-latest",
            "aliyun/qwen-vl-ocr-latest",
            "aliyun/qvq-max-latest",
            "aliyun/qwen2.5-vl-32b-instruct",
            "aliyun/qwen2.5-vl-72b-instruct",
            "openai/o3",
            "openai/o4-mini-deep-research",
            "openai/gpt-4.1-nano",
            "openai/o4-mini",
            "openai/o1",
            "openai/gpt-4o-mini",
            "openai/codex-mini-latest",
            "openai/gpt-4o-audio-preview",
            "openai/gpt-4.1-mini",
            "openai/gpt-image-1",
            "openai/gpt-4o",
            "openai/gpt-4.1",
            "anthropic/claude-sonnet-4",
            "anthropic/claude-sonnet-4-thinking",
            "anthropic/claude-opus-4-thinking",
            "anthropic/claude-3-5-haiku",
            "anthropic/claude-opus-4",
            "ark/doubao-vision-pro",
        ]

        allowed_models = [
        # openai_models
        "gpt-41",
        "gpt-45",
        "gpt-4o",
        "o1",
        "o3",
        "o4",
        "gpt-4",
        "gpt-35",
        "r1",
        # google_models
        "gemini-1.5-flash",
        "gemini-1.5-pro",
        "gemini-2.0-flash",
        "gemini-2.5-pro",
        # anthropic_models
        "claude-3-haiku",
        "claude-3-sonnet",
        "claude-3-opus",
        "claude-3-5-haiku",
        "claude-3-5-sonnet",
        "claude-3-7-sonnet",
        # mistral_models
        "codestral",
        "open-codestral-mamba",
        "mistral",
        "ministral",
        "pixtral",
        # unknown
        "unknown"
        ]

        hepai_allowed_models = [
            "aliyun/",
            "openai/",
            "anthropic/",
            "ark/",
            "deepseek-ai/",
            "hepai/",
        ]
        model = kwargs.get("model", "")
        all_allowed_models = allowed_models+hepai_allowed_models
        for allowed_model in all_allowed_models:
            if allowed_model in model.lower():
                if allowed_model in hepai_allowed_models:
                    if model in self.r1_series:
                        kwargs["model_info"]["family"] = ModelFamily.R1
                    elif model in self.v3_series:
                        kwargs["model_info"]["family"] = ModelFamily.GPT_4O
                else:
                    kwargs["model_info"]["family"] = allowed_model
                
                if model in self.vsion_series:
                    kwargs["model_info"]["vision"] = True
                
                kwargs["model_info"]["function_calling"] = True
                kwargs["model_info"]["json_output"] = True
                kwargs["model_info"]["structured_output"] = True
                break

        super().__init__(**kwargs)
    
    async def create_stream(
        self,
        messages: Sequence[LLMMessage],
        *,
        tools: Sequence[Tool | ToolSchema] = [],
        json_output: Optional[bool | type[BaseModel]] = None,
        extra_create_args: Mapping[str, Any] = {},
        cancellation_token: Optional[CancellationToken] = None,
        max_consecutive_empty_chunk_tolerance: int = 0,
    ) -> AsyncGenerator[Union[str, CreateResult], None]:
        """Create a stream of string chunks from the model ending with a :class:`~autogen_core.models.CreateResult`.

        Extends :meth:`autogen_core.models.ChatCompletionClient.create_stream` to support OpenAI API.

        In streaming, the default behaviour is not return token usage counts.
        See: `OpenAI API reference for possible args <https://platform.openai.com/docs/api-reference/chat/create>`_.

        You can set `extra_create_args={"stream_options": {"include_usage": True}}`
        (if supported by the accessed API) to
        return a final chunk with usage set to a :class:`~autogen_core.models.RequestUsage` object
        with prompt and completion token counts,
        all preceding chunks will have usage as `None`.
        See: `OpenAI API reference for stream options <https://platform.openai.com/docs/api-reference/chat/create#chat-create-stream_options>`_.

        Other examples of supported arguments that can be included in `extra_create_args`:
            - `temperature` (float): Controls the randomness of the output. Higher values (e.g., 0.8) make the output more random, while lower values (e.g., 0.2) make it more focused and deterministic.
            - `max_tokens` (int): The maximum number of tokens to generate in the completion.
            - `top_p` (float): An alternative to sampling with temperature, called nucleus sampling, where the model considers the results of the tokens with top_p probability mass.
            - `frequency_penalty` (float): A value between -2.0 and 2.0 that penalizes new tokens based on their existing frequency in the text so far, decreasing the likelihood of repeated phrases.
            - `presence_penalty` (float): A value between -2.0 and 2.0 that penalizes new tokens based on whether they appear in the text so far, encouraging the model to talk about new topics.
        """

        create_params = self._process_create_args(
            messages,
            tools,
            json_output,
            extra_create_args,
        )

        # # # TODO: we should remove tool_calls?
        # new_oai_messages = []
        # for message in create_params.messages:
        #     # if message.get("role") == "tool":
        #         # message["role"] = "assistant"
        #         # message.pop("tool_call_id")
        #         # continue
        #     if message.get("tool_calls") is not None:
        #         # tool_calls = message.pop("tool_calls")
        #         # if message.get("content") is None:
        #         #     message["content"] = ""
        #         # message["content"] += str(tool_calls)
        #         tool_call_messages = []
        #         # content = message.get("content")
        #         # role = "tool"
        #         tool_calls = message.get("tool_calls")
        #         for tool_call in tool_calls:
        #             tool_call_id = tool_call.get("id")
        #             tool_call_messages.append(
        #                 {
        #                     "role": "tool",
        #                     "tool_call_id": tool_call_id,
        #                     "content": str(tool_call),
        #                 }
        #             )

        #     new_oai_messages.append(message)
        # create_params.messages = new_oai_messages

        if max_consecutive_empty_chunk_tolerance != 0:
            warnings.warn(
                "The 'max_consecutive_empty_chunk_tolerance' parameter is deprecated and will be removed in the future releases. All of empty chunks will be skipped with a warning.",
                DeprecationWarning,
                stacklevel=2,
            )

        if create_params.response_format is not None:
            chunks = self._create_stream_chunks_beta_client(
                tool_params=create_params.tools,
                oai_messages=create_params.messages,
                response_format=create_params.response_format,
                create_args_no_response_format=create_params.create_args,
                cancellation_token=cancellation_token,
            )
        else:
            chunks = self._create_stream_chunks(
                tool_params=create_params.tools,
                oai_messages=create_params.messages,
                create_args=create_params.create_args,
                cancellation_token=cancellation_token,
            )

        # Prepare data to process streaming chunks.
        chunk: ChatCompletionChunk | None = None
        stop_reason = None
        maybe_model = None
        content_deltas: List[str] = []
        thought_deltas: List[str] = []
        full_tool_calls: Dict[int, FunctionCall] = {}
        logprobs: Optional[List[ChatCompletionTokenLogprob]] = None

        empty_chunk_warning_has_been_issued: bool = False
        empty_chunk_warning_threshold: int = 10
        empty_chunk_count = 0
        first_chunk = True
        is_reasoning = False

        # Process the stream of chunks.
        async for chunk in chunks:
            if first_chunk:
                first_chunk = False
                # Emit the start event.
                logger.info(
                    LLMStreamStartEvent(
                        messages=cast(List[Dict[str, Any]], create_params.messages),
                    )
                )

            # Set the model from the lastest chunk.
            maybe_model = chunk.model

            # Empty chunks has been observed when the endpoint is under heavy load.
            #  https://github.com/microsoft/autogen/issues/4213
            if len(chunk.choices) == 0:
                empty_chunk_count += 1
                if not empty_chunk_warning_has_been_issued and empty_chunk_count >= empty_chunk_warning_threshold:
                    empty_chunk_warning_has_been_issued = True
                    warnings.warn(
                        f"Received more than {empty_chunk_warning_threshold} consecutive empty chunks. Empty chunks are being ignored.",
                        stacklevel=2,
                    )
                continue
            else:
                empty_chunk_count = 0

            if len(chunk.choices) > 1:
                # This is a multi-choice chunk, we need to warn the user.
                warnings.warn(
                    f"Received a chunk with {len(chunk.choices)} choices. Only the first choice will be used.",
                    UserWarning,
                    stacklevel=2,
                )

            # Set the choice to the first choice in the chunk.
            choice = chunk.choices[0]

            # for liteLLM chunk usage, do the following hack keeping the pervious chunk.stop_reason (if set).
            # set the stop_reason for the usage chunk to the prior stop_reason
            stop_reason = choice.finish_reason if chunk.usage is None and stop_reason is None else stop_reason
            maybe_model = chunk.model

            reasoning_content: str | None = None
            if choice.delta.model_extra is not None and "reasoning_content" in choice.delta.model_extra:
                # If there is a reasoning_content field, then we populate the thought field. This is for models such as R1.
                reasoning_content = choice.delta.model_extra.get("reasoning_content")

            if isinstance(reasoning_content, str) and len(reasoning_content) > 0:
                if not is_reasoning:
                    # Enter reasoning mode.
                    reasoning_content = "<think>" + reasoning_content
                    is_reasoning = True
                thought_deltas.append(reasoning_content)
                yield reasoning_content
            elif is_reasoning:
                # Exit reasoning mode.
                reasoning_content = "</think>"
                thought_deltas.append(reasoning_content)
                is_reasoning = False
                yield reasoning_content

            # First try get content
            if choice.delta.content:
                content_deltas.append(choice.delta.content)
                if len(choice.delta.content) > 0:
                    yield choice.delta.content
                # NOTE: for OpenAI, tool_calls and content are mutually exclusive it seems, so we can skip the rest of the loop.
                # However, this may not be the case for other APIs -- we should expect this may need to be updated.
                continue
            # Otherwise, get tool calls
            if choice.delta.tool_calls is not None:
                for tool_call_chunk in choice.delta.tool_calls:
                    idx = tool_call_chunk.index
                    if idx not in full_tool_calls:
                        # We ignore the type hint here because we want to fill in type when the delta provides it
                        full_tool_calls[idx] = FunctionCall(id="", arguments="", name="")

                    if tool_call_chunk.id is not None:
                        full_tool_calls[idx].id += tool_call_chunk.id

                    if tool_call_chunk.function is not None:
                        if tool_call_chunk.function.name is not None:
                            full_tool_calls[idx].name += tool_call_chunk.function.name
                        if tool_call_chunk.function.arguments is not None:
                            full_tool_calls[idx].arguments += tool_call_chunk.function.arguments
            if choice.logprobs and choice.logprobs.content:
                logprobs = [
                    ChatCompletionTokenLogprob(
                        token=x.token,
                        logprob=x.logprob,
                        top_logprobs=[TopLogprob(logprob=y.logprob, bytes=y.bytes) for y in x.top_logprobs],
                        bytes=x.bytes,
                    )
                    for x in choice.logprobs.content
                ]

        # Finalize the CreateResult.

        # TODO: can we remove this?
        if stop_reason == "function_call":
            raise ValueError("Function calls are not supported in this context")

        # We need to get the model from the last chunk, if available.
        model = maybe_model or create_params.create_args["model"]
        model = model.replace("gpt-35", "gpt-3.5")  # hack for Azure API

        # Because the usage chunk is not guaranteed to be the last chunk, we need to check if it is available.
        if chunk and chunk.usage:
            prompt_tokens = chunk.usage.prompt_tokens
            completion_tokens = chunk.usage.completion_tokens
        else:
            prompt_tokens = 0
            completion_tokens = 0
        usage = RequestUsage(
            prompt_tokens=prompt_tokens,
            completion_tokens=completion_tokens,
        )

        # Detect whether it is a function call or just text.
        content: Union[str, List[FunctionCall]]
        thought: str | None = None
        # Determine the content and thought based on what was collected
        if full_tool_calls:
            # This is a tool call response
            content = list(full_tool_calls.values())
            if thought_deltas:
                thought = "".join(thought_deltas).lstrip("<think>").rstrip("</think>")
            else:
            # Store any text alongside tool calls as thoughts
                thought = "".join(content_deltas)
        else:
            # This is a text response (possibly with thoughts)
            if content_deltas:
                content = "".join(content_deltas)
            else:
                warnings.warn(
                    "No text content or tool calls are available. Model returned empty result.",
                    stacklevel=2,
                )
                content = ""

            # Set thoughts if we have any reasoning content.
            if thought_deltas:
                thought = "".join(thought_deltas).lstrip("<think>").rstrip("</think>")

            # This is for local R1 models whose reasoning content is within the content string.
            if isinstance(content, str) and self._model_info["family"] == ModelFamily.R1 and thought is None:
                thought, content = parse_r1_content(content)

        # Create the result.
        result = CreateResult(
            finish_reason=normalize_stop_reason(stop_reason),
            content=content,
            usage=usage,
            cached=False,
            logprobs=logprobs,
            thought=thought,
        )

        # Log the end of the stream.
        logger.info(
            LLMStreamEndEvent(
                response=result.model_dump(),
                prompt_tokens=usage.prompt_tokens,
                completion_tokens=usage.completion_tokens,
            )
        )

        # Update the total usage.
        self._total_usage = _add_usage(self._total_usage, usage)
        self._actual_usage = _add_usage(self._actual_usage, usage)

        # Yield the CreateResult.
        yield result
    
    def count_tokens(self, messages: Sequence[LLMMessage], *, tools: Sequence[Tool | ToolSchema] = []) -> int:

    
        model_name = self._model_info.get("token_model")
        if model_name is None:
            raise ValueError(f"Maybe you can provide a token_model with in {_MODEL_TOKEN_LIMITS} when provide the model_info")

        return count_tokens_openai(
            messages,
            model_name,
            add_name_prefixes=self._add_name_prefixes,
            tools=tools,
            model_family=self._model_info["family"],
        )

    def remaining_tokens(
            self, 
            messages: Sequence[LLMMessage], 
            *, 
            tools: Sequence[Tool | ToolSchema] = [],
            token_limit: int | None = None,
            ) -> int:
        model_name = self._model_info.get("token_model")
        if model_name is None:
            raise ValueError(f"Maybe you can provide a token_model with in {_MODEL_TOKEN_LIMITS} when provide the model_info")
        
        if token_limit is None:
            token_limit = _model_info.get_token_limit(model_name)
        return token_limit - self.count_tokens(messages, tools=tools)



import asyncio
import json
import logging
from typing import (
    Any,
    AsyncGenerator,
    Coroutine,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Union,
    cast,
)

from anthropic import  AsyncStream
from anthropic.types import (
    # Base64ImageSourceParam,
    # ContentBlock,
    # ImageBlockParam,
    # Message,
    MessageParam,
    RawMessageStreamEvent,  # type: ignore
    TextBlock,
    # TextBlockParam,
    # ToolParam,
    # ToolResultBlockParam,
    ToolUseBlock,
)
from autogen_core import (
    EVENT_LOGGER_NAME,
    TRACE_LOGGER_NAME,
    CancellationToken,
    Component,
    FunctionCall,
    Image,
)
from autogen_core.logging import LLMCallEvent, LLMStreamEndEvent, LLMStreamStartEvent
from autogen_core.models import (
    AssistantMessage,
    ChatCompletionClient,
    CreateResult,
    FinishReasons,
    FunctionExecutionResultMessage,
    LLMMessage,
    ModelCapabilities,  # type: ignore
    ModelInfo,
    RequestUsage,
    SystemMessage,
    UserMessage,
    validate_model_info,
)
from autogen_core.tools import Tool, ToolSchema
from pydantic import BaseModel

logger = logging.getLogger(EVENT_LOGGER_NAME)


from autogen_ext.models.anthropic import AnthropicChatCompletionClient, AnthropicBedrockClientConfiguration
from autogen_ext.models.anthropic._anthropic_client import (
    to_anthropic_type, 
    convert_tools, 
    normalize_name,
    normalize_stop_reason,
    _add_usage
    )

class HepAIAnthropicChatCompletionClient(AnthropicChatCompletionClient):

    async def create_stream_tmp(
        self,
        messages: Sequence[LLMMessage],
        *,
        tools: Sequence[Tool | ToolSchema] = [],
        json_output: Optional[bool | type[BaseModel]] = None,
        extra_create_args: Mapping[str, Any] = {},
        cancellation_token: Optional[CancellationToken] = None,
        max_consecutive_empty_chunk_tolerance: int = 0,
    ) -> AsyncGenerator[Union[str, CreateResult], None]:
        """
        Creates an AsyncGenerator that yields a stream of completions based on the provided messages and tools.
        """
        # Copy create args and update with extra args
        create_args = self._create_args.copy()
        create_args.update(extra_create_args)

        # Check for vision capability if images are present
        if self.model_info["vision"] is False:
            for message in messages:
                if isinstance(message, UserMessage):
                    if isinstance(message.content, list) and any(isinstance(x, Image) for x in message.content):
                        raise ValueError("Model does not support vision and image was provided")

        # Handle JSON output format
        if json_output is not None:
            if self.model_info["json_output"] is False and json_output is True:
                raise ValueError("Model does not support JSON output")

            if json_output is True:
                create_args["response_format"] = {"type": "json_object"}

            if isinstance(json_output, type):
                raise ValueError("Structured output is currently not supported for Anthropic models")

        # Process system message separately
        system_message = None
        anthropic_messages: List[MessageParam] = []

        # Merge continuous system messages into a single message
        messages = self._merge_system_messages(messages)
        messages = self._rstrip_last_assistant_message(messages)

        # NOTE：Not support ToolExecuteMessage
        for message in messages:
            if isinstance(message, SystemMessage):
                if system_message is not None:
                    # if that case, system message is must only one
                    raise ValueError("Multiple system messages are not supported")
                system_message = to_anthropic_type(message)
            else:
                anthropic_message = to_anthropic_type(message)
                if isinstance(anthropic_message, list):
                    anthropic_messages.extend(anthropic_message)
                elif isinstance(anthropic_message["content"], list):
                    anthropic_message_new = []
                    for message_i in anthropic_message["content"]:
                        if isinstance(message_i, TextBlock):
                            anthropic_message_new.append({"type": "text", "text": message_i.text})
                        elif isinstance(message_i, ToolUseBlock):
                            anthropic_message_new.append({
                                "type": "tool_use",
                                "id": message_i.id,
                                "name": message_i.name,
                                "input": message_i.input,
                            })
                        elif isinstance(message_i, BaseModel):
                            anthropic_message_new.append(message_i.model_dump())
                        elif isinstance(message_i, dict):
                            anthropic_message_new.append(message_i)
                    anthropic_message["content"] = anthropic_message_new
                    anthropic_messages.append(anthropic_message)
                elif isinstance(anthropic_message, str):
                    msg = MessageParam(
                        role="user" if isinstance(message, UserMessage) else "assistant", content=anthropic_message
                    )
                    anthropic_messages.append(msg)
                else:
                    anthropic_messages.append(anthropic_message)

        # TODO: can't convert ToolCallMessage to text
        # for message in messages:
        #     if isinstance(message, SystemMessage):
        #         if system_message is not None:
        #             # if that case, system message is must only one
        #             raise ValueError("Multiple system messages are not supported")
        #         system_message = to_anthropic_type(message)
        #     else:
        #         anthropic_message = to_anthropic_type(message)

        #         if isinstance(anthropic_message, list):
        #             # anthropic_messages.extend(anthropic_message)
        #             # For ToolExecuteMessage
        #             for message_i in anthropic_message:
        #                 if isinstance(message_i["content"], list):
        #                     new_content = ""
        #                     for content_i in message_i["content"]:
        #                         if isinstance(content_i, dict):
        #                             new_content += content_i["content"]+"\n"
        #                     message_i["content"] = new_content
        #                 anthropic_messages.append(message_i)
        #         elif isinstance(anthropic_message, str):
        #             msg = MessageParam(
        #                 role="user" if isinstance(message, UserMessage) else "assistant", content=anthropic_message
        #             )
        #             anthropic_messages.append(msg)
        #         elif isinstance(anthropic_message["content"], list):
        #             new_content = ""
        #             for item in anthropic_message["content"]:
        #                 if isinstance(item, TextBlock):
        #                     new_content += item.text+"\n"
        #                 elif isinstance(item, ToolUseBlock):
        #                     # new_content += f"tool_name: {item.name}, arguments: {str(item.input)}"
        #                     new_content += item.model_dump_json()
        #                 elif isinstance(item, dict):
        #                     # new_content += item["content"]
        #                     new_content += str(item)
        #                 else:
        #                     continue
        #             anthropic_message["content"] = new_content
        #             anthropic_messages.append(anthropic_message)
        #         else:
        #             anthropic_messages.append(anthropic_message)

        # Check for function calling support
        if self.model_info["function_calling"] is False and len(tools) > 0:
            raise ValueError("Model does not support function calling")

        # Set up the request
        request_args: Dict[str, Any] = {
            "model": create_args["model"],
            "messages": anthropic_messages,
            "max_tokens": create_args.get("max_tokens", 4096),
            "temperature": create_args.get("temperature", 1.0),
            "stream": True,
        }

        # Add system message if present
        if system_message is not None:
            request_args["system"] = system_message

        # Check if any message is a tool result
        has_tool_results = any(isinstance(msg, FunctionExecutionResultMessage) for msg in messages)

        # Add tools if present
        if len(tools) > 0:
            converted_tools = convert_tools(tools)
            self._last_used_tools = converted_tools
            request_args["tools"] = converted_tools
        elif has_tool_results:
            request_args["tools"] = self._last_used_tools

        # Optional parameters
        for param in ["top_p", "top_k", "stop_sequences", "metadata"]:
            if param in create_args:
                request_args[param] = create_args[param]

        # Stream the response
        stream_future: asyncio.Task[AsyncStream[RawMessageStreamEvent]] = asyncio.ensure_future(
            cast(Coroutine[Any, Any, AsyncStream[RawMessageStreamEvent]], self._client.messages.create(**request_args))
        )

        if cancellation_token is not None:
            cancellation_token.link_future(stream_future)  # type: ignore

        stream: AsyncStream[RawMessageStreamEvent] = cast(AsyncStream[RawMessageStreamEvent], await stream_future)  # type: ignore

        text_content: List[str] = []
        tool_calls: Dict[str, Dict[str, Any]] = {}  # Track tool calls by ID
        current_tool_id: Optional[str] = None
        input_tokens: int = 0
        output_tokens: int = 0
        stop_reason: Optional[str] = None

        first_chunk = True
        serialized_messages: List[Dict[str, Any]] = [self._serialize_message(msg) for msg in anthropic_messages]

        # Process the stream
        async for chunk in stream:
            if first_chunk:
                first_chunk = False
                # Emit the start event.
                logger.info(
                    LLMStreamStartEvent(
                        messages=serialized_messages,
                    )
                )
            # Handle different event types
            if chunk.type == "content_block_start":
                if chunk.content_block.type == "tool_use":
                    # Start of a tool use block
                    current_tool_id = chunk.content_block.id
                    tool_calls[current_tool_id] = {
                        "id": chunk.content_block.id,
                        "name": chunk.content_block.name,
                        "input": "",  # Will be populated from deltas
                    }

            elif chunk.type == "content_block_delta":
                if hasattr(chunk.delta, "type") and chunk.delta.type == "text_delta":
                    # Handle text content
                    delta_text = chunk.delta.text
                    text_content.append(delta_text)
                    if delta_text:
                        yield delta_text

                # Handle tool input deltas - they come as InputJSONDelta
                elif hasattr(chunk.delta, "type") and chunk.delta.type == "input_json_delta":
                    if current_tool_id is not None and hasattr(chunk.delta, "partial_json"):
                        # Accumulate partial JSON for the current tool
                        tool_calls[current_tool_id]["input"] += chunk.delta.partial_json

            elif chunk.type == "content_block_stop":
                # End of a content block (could be text or tool)
                current_tool_id = None

            elif chunk.type == "message_delta":
                if hasattr(chunk.delta, "stop_reason") and chunk.delta.stop_reason:
                    stop_reason = chunk.delta.stop_reason

                # Get usage info if available
                if hasattr(chunk, "usage") and hasattr(chunk.usage, "output_tokens"):
                    output_tokens = chunk.usage.output_tokens

            elif chunk.type == "message_start":
                if hasattr(chunk, "message") and hasattr(chunk.message, "usage"):
                    if hasattr(chunk.message.usage, "input_tokens"):
                        input_tokens = chunk.message.usage.input_tokens
                    if hasattr(chunk.message.usage, "output_tokens"):
                        output_tokens = chunk.message.usage.output_tokens

        # Prepare the final response
        usage = RequestUsage(
            prompt_tokens=input_tokens,
            completion_tokens=output_tokens,
        )

        # Determine content based on what was received
        content: Union[str, List[FunctionCall]]
        thought = None

        if tool_calls:
            # We received tool calls
            if text_content:
                # Text before tool calls is treated as thought
                thought = "".join(text_content)

            # Convert tool calls to FunctionCall objects
            content = []
            for _, tool_data in tool_calls.items():
                # Parse the JSON input if needed
                input_str = tool_data["input"]
                try:
                    # If it's valid JSON, parse it; otherwise use as-is
                    if input_str.strip().startswith("{") and input_str.strip().endswith("}"):
                        parsed_input = json.loads(input_str)
                        input_str = json.dumps(parsed_input)  # Re-serialize to ensure valid JSON
                except json.JSONDecodeError:
                    # Keep as string if not valid JSON
                    pass

                content.append(
                    FunctionCall(
                        id=tool_data["id"],
                        name=normalize_name(tool_data["name"]),
                        arguments=input_str,
                    )
                )
        else:
            # Just text content
            content = "".join(text_content)

        # Create the final result
        result = CreateResult(
            finish_reason=normalize_stop_reason(stop_reason),
            content=content,
            usage=usage,
            cached=False,
            thought=thought,
        )

        # Emit the end event.
        logger.info(
            LLMStreamEndEvent(
                response=result.model_dump(),
                prompt_tokens=usage.prompt_tokens,
                completion_tokens=usage.completion_tokens,
            )
        )

        # Update usage statistics
        self._total_usage = _add_usage(self._total_usage, usage)
        self._actual_usage = _add_usage(self._actual_usage, usage)

        yield result


# defines how core data types are serialized and stored in the database

from datetime import datetime
from enum import Enum
from typing import Any, List, Optional, Union, Dict

from autogen_core import ComponentModel
from pydantic import field_serializer
from sqlalchemy import ForeignKey, Integer
from sqlmodel import JSON, Column, DateTime, Field, SQLModel, func

from .types import (
    GalleryConfig,
    MessageConfig,
    MessageMeta,
    SettingsConfig,
    TeamResult,
    GalleryComponents,
    GalleryMetadata,
)

class RunStatus(str, Enum):
    CREATED = "created"
    ACTIVE = "active"
    COMPLETE = "complete"
    ERROR = "error"
    STOPPED = "stopped"
    AWAITING_INPUT = "awaiting_input"
    PAUSED = "paused"

class AutoGenMessage(SQLModel, table=True):
    __table_args__ = {"sqlite_autoincrement": True}
    id: Optional[int] = Field(default=None, primary_key=True)
    created_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now()),
    )  # pylint: disable=not-callable
    updated_at: datetime = Field(
        default_factory=datetime.now,
        sa_column=Column(DateTime(timezone=True), onupdate=func.now()),
    )  # pylint: disable=not-callable
    version: Optional[str] = "0.0.1"
    # task info
    user_id: Optional[str] = None # email or other unique identifier
    thread_id: Optional[str] = None # unique identifier for the conversation thread
    # message content
    config: Union[MessageConfig, dict[str, Any]] = Field(
        default_factory=lambda: MessageConfig(source="", content=""),
        sa_column=Column(JSON),
    )
    message_meta: Optional[Union[MessageMeta, dict[str, Any]]] = Field(
        default={}, sa_column=Column(JSON)
    )

    @field_serializer("created_at", "updated_at")
    def serialize_datetime(cls, value: datetime) -> str:
        if isinstance(value, datetime):
            return value.isoformat()

class UserInput(SQLModel, table=True):
    __table_args__ = {"sqlite_autoincrement": True}
    id: Optional[int] = Field(default=None, primary_key=True)
    created_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now()),
    )  # pylint: disable=not-callable
    updated_at: datetime = Field(
        default_factory=datetime.now,
        sa_column=Column(DateTime(timezone=True), onupdate=func.now()),
    )  # pylint: disable=not-callable
    version: Optional[str] = "0.0.1"
    # task info
    user_id: Optional[str] = None # email or other unique identifier
    thread_id: Optional[str] = None # unique identifier for the conversation thread
    # user input
    user_messages: Union[List[AutoGenMessage], List[dict[str, Any]]] = Field(
        default_factory=list, sa_column=Column(JSON)
    )
    user_last_message: Optional[Dict[str, Any]] = Field(
        default_factory=dict, sa_column=Column(JSON)
    )
    api_key: Optional[str] = None # user API key for authentication
    temperature: Optional[float] = None
    max_tokens: Optional[int] = None
    top_p: Optional[float] = None
    cache_seed: Optional[int] = None
    n: Optional[int] = None
    stream: Optional[bool] = True
    extra_requests: Optional[Dict[str, Any]] = Field(
        default_factory=dict, sa_column=Column(JSON)
    )

    @field_serializer("created_at", "updated_at")
    def serialize_datetime(cls, value: datetime) -> str:
        if isinstance(value, datetime):
            return value.isoformat()
        
class SingleTask(SQLModel, table=True):
    __table_args__ = {"sqlite_autoincrement": True}
    id: Optional[int] = Field(default=None, primary_key=True)
    created_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now()),
    )  # pylint: disable=not-callable
    updated_at: datetime = Field(
        default_factory=datetime.now,
        sa_column=Column(DateTime(timezone=True), onupdate=func.now()),
    )  # pylint: disable=not-callable
    version: Optional[str] = "0.0.1"
    # user info
    user_id: Optional[str] = None
    thread_id: Optional[str] = None 
    # task info
    task_id: Optional[str] = None
    is_node_task: bool = False
    node_task_id: Optional[str] = None
    previous_id: Optional[str] = None
    next_id: Optional[str] = None
    # task content
    content: Optional[str] = None # for dict input, please json.dumps
    files: Optional[dict[str, Any]] = Field(
        default_factory=dict, sa_column=Column(JSON)
    )
    task_messages: Union[List[AutoGenMessage], List[dict[str, Any]]] = Field(
        default_factory=list, sa_column=Column(JSON)
    )
    meta: Optional[Dict[str, Any]] =  Field(
        default_factory=dict, sa_column=Column(JSON)
    )

    @field_serializer("created_at", "updated_at")
    def serialize_datetime(cls, value: datetime) -> str:
        if isinstance(value, datetime):
            return value.isoformat()

class Tasks(SQLModel, table=True):
    __table_args__ = {"sqlite_autoincrement": True}
    id: Optional[int] = Field(default=None, primary_key=True)
    created_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now()),
    )  # pylint: disable=not-callable
    updated_at: datetime = Field(
        default_factory=datetime.now,
        sa_column=Column(DateTime(timezone=True), onupdate=func.now()),
    )  # pylint: disable=not-callable
    version: Optional[str] = "0.0.1"
    # user info
    user_id: Optional[str] = None
    thread_id: Optional[str] = None 
    # task info
    tasks_id: Optional[str] = None
    # task content
    content: Optional[str] = None # for dict input, please json.dumps
    files: Optional[dict[str, Any]] = Field(
        default_factory=dict, sa_column=Column(JSON)
    )
    # sub tasks
    tasks: Optional[str] = None # for dict input, please json.dumps(SingleTask)

    meta: Optional[Dict[str, Any]] =  Field(
        default_factory=dict, sa_column=Column(JSON)
    )

    @field_serializer("created_at", "updated_at")
    def serialize_datetime(cls, value: datetime) -> str:
        if isinstance(value, datetime):
            return value.isoformat()
        
class PlanCheck(SQLModel, table=True):
    __table_args__ = {"sqlite_autoincrement": True}
    id: Optional[int] = Field(default=None, primary_key=True)
    created_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now()),
    )  # pylint: disable=not-callable
    updated_at: datetime = Field(
        default_factory=datetime.now,
        sa_column=Column(DateTime(timezone=True), onupdate=func.now()),
    )  # pylint: disable=not-callable
    version: Optional[str] = "0.0.1"
    # user info
    user_id: Optional[str] = None
    thread_id: Optional[str] = None 
    # task info
    task: Optional[str] = None
    steps: Optional[List[dict[str, Any]]] = Field(
        default_factory=list, sa_column=Column(JSON)
    )
    session_id: Optional[int] = None

    @field_serializer("created_at", "updated_at")
    def serialize_datetime(cls, value: datetime) -> str:
        if isinstance(value, datetime):
            return value.isoformat()
        
class Thread(SQLModel, table=True):
    __table_args__ = {"sqlite_autoincrement": True}
    id: Optional[int] = Field(default=None, primary_key=True)
    created_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now()),
    )  # pylint: disable=not-callable
    updated_at: datetime = Field(
        default_factory=datetime.now,
        sa_column=Column(DateTime(timezone=True), onupdate=func.now()),
    )  # pylint: disable=not-callable
    version: Optional[str] = "0.0.1"

    # uesr info
    user_id: Optional[str] = None # email or other unique identifier
    thread_id: Optional[str] = None # unique identifier for the conversation thread
    user_input: Optional[dict[str, Any]] = Field(
        default=None, sa_column=Column(JSON)
    )
    
    # run info
    status: RunStatus = Field(default=RunStatus.CREATED)

    # messages
    messages: Union[List[AutoGenMessage], List[dict[str, Any]]] = Field(
        default_factory=list, sa_column=Column(JSON)
    )
    error_message: Optional[str] = None

    # Store the userproxy input for the current task
    input_request: Optional[dict[str, Any]] = Field(
        default=None, sa_column=Column(JSON)
    )

    # task info
    tasks: Union[Tasks, List[dict[str, Any]]] = Field(
        default_factory=list, sa_column=Column(JSON)
    )

    # Store TeamResult which contains TaskResult
    team_result: Union[TeamResult, dict[str, Any]] = Field(
        default=None, sa_column=Column(JSON)
    )

    # save state of Agent/Group system by base64 encoded string
    state: Optional[str] = None

    meta: Optional[Dict[str, Any]] =  Field(
        default_factory=dict, sa_column=Column(JSON)
    )

    @field_serializer("created_at", "updated_at")
    def serialize_datetime(cls, value: datetime) -> str:
        if isinstance(value, datetime):
            return value.isoformat()

class AgentJson(SQLModel, table=True):
    __table_args__ = {"sqlite_autoincrement": True}
    id: Optional[int] = Field(default=None, primary_key=True)
    created_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now()),
    )  # pylint: disable=not-callable
    updated_at: datetime = Field(
        default_factory=datetime.now,
        sa_column=Column(DateTime(timezone=True), onupdate=func.now()),
    )  # pylint: disable=not-callable
    user_id: Optional[str] = None
    version: Optional[str] = "0.0.1"
    component: Union[ComponentModel, dict[str, Any]] = Field(sa_column=Column(JSON))

    @field_serializer("created_at", "updated_at")
    def serialize_datetime(cls, value: datetime) -> str:
        if isinstance(value, datetime):
            return value.isoformat()


class InputType(str, Enum):
    TEXT_INPUT = "text_input"
    APPROVAL = "approval"


# class Run(SQLModel, table=True):
#     """Represents a single execution run within a session"""

#     __table_args__ = {"sqlite_autoincrement": True}

#     id: Optional[int] = Field(default=None, primary_key=True)
#     created_at: datetime = Field(
#         sa_column=Column(DateTime(timezone=True), server_default=func.now()),
#     )
#     updated_at: datetime = Field(
#         default_factory=datetime.now,
#         sa_column=Column(DateTime(timezone=True), onupdate=func.now()),
#     )
#     session_id: Optional[int] = Field(
#         default=None,
#         sa_column=Column(
#             Integer, ForeignKey("session.id", ondelete="CASCADE"), nullable=False
#         ),
#     )
#     status: RunStatus = Field(default=RunStatus.CREATED)

#     # Store the original user task
#     task: Union[MessageConfig, dict[str, Any]] = Field(
#         default_factory=lambda: MessageConfig(source="", content=""),
#         sa_column=Column(JSON),
#     )

#     # Store TeamResult which contains TaskResult
#     team_result: Union[TeamResult, dict[str, Any]] = Field(
#         default=None, sa_column=Column(JSON)
#     )

#     error_message: Optional[str] = None
#     version: Optional[str] = "0.0.1"
#     messages: Union[List[Message], List[dict[str, Any]]] = Field(
#         default_factory=list, sa_column=Column(JSON)
#     )

#     user_id: Optional[str] = None
#     state: Optional[str] = None

#     input_request: Optional[dict[str, Any]] = Field(
#         default=None, sa_column=Column(JSON)
#     )

#     @field_serializer("created_at", "updated_at")
#     def serialize_datetime(cls, value: datetime) -> str:
#         if isinstance(value, datetime):
#             return value.isoformat()



# class Plan(SQLModel, table=True):
#     __table_args__ = {"sqlite_autoincrement": True}
#     id: Optional[int] = Field(default=None, primary_key=True)
#     created_at: datetime = Field(
#         sa_column=Column(DateTime(timezone=True), server_default=func.now()),
#     )  # pylint: disable=not-callable
#     updated_at: datetime = Field(
#         default_factory=datetime.now,
#         sa_column=Column(DateTime(timezone=True), onupdate=func.now()),
#     )  # pylint: disable=not-callable
#     user_id: Optional[str] = None
#     version: Optional[str] = "0.0.1"
#     task: Optional[str] = None
#     steps: Optional[List[dict[str, Any]]] = Field(
#         default_factory=list, sa_column=Column(JSON)
#     )
#     session_id: Optional[int] = None

#     @field_serializer("created_at", "updated_at")
#     def serialize_datetime(cls, value: datetime) -> str:
#         if isinstance(value, datetime):
#             return value.isoformat()



##

DatabaseModel = AutoGenMessage | UserInput | SingleTask | Tasks | PlanCheck | Thread | AgentJson



from pathlib import Path
import re
from pydantic import BaseModel
from drsai.modules.components import ComponentBase, Component

class SkillLoaderConfig(BaseModel):
    """Configuration for SkillLoader."""
    skills_dir: str | list[str]

class SkillLoader(Component[SkillLoaderConfig], ComponentBase[BaseModel]):
    """
    Loads and manages skills from SKILL.md files.

    A skill is a FOLDER containing:
    - SKILL.md (required): YAML frontmatter + markdown instructions
    - scripts/ (optional): Helper scripts the model can run
    - references/ (optional): Additional documentation
    - assets/ (optional): Templates, files for output

    SKILL.md Format:
    ----------------
        ---
        name: pdf
        description: Process PDF files. Use when reading, creating, or merging PDFs.
        compatibility: pdftotext, pypdf
        ---

        # PDF Processing Skill

        ## Reading PDFs

        Use pdftotext for quick extraction:
        ```bash
        pdftotext input.pdf -
        ```
        ...

    The YAML frontmatter provides metadata (name, description, compatibility).
    The markdown body provides detailed instructions.
    """

    component_config_schema = SkillLoaderConfig
    component_provider_override = "drsai.modules.components.skills.SkillLoader"
    component_type =  "skill_loader"

    def __init__(self, skills_dir: str | list[str],):

        self.skills = {}
        self.add_skills_by_dir(skills_dir=skills_dir)

    def safe_path(self, path: str) -> Path:
        """Make sure path is safe and exists."""
        path_obj = Path(path)
        resolved_path = path_obj.resolve()
        # if ".." in path.split("/") or ".." in path.split("\\"):
        #     try:
        #         resolved_path.relative_to(Path.cwd())
        #     except ValueError:
        #         raise ValueError(f"Unsafe path detected: {path}")
        if not resolved_path.exists():
            raise ValueError(f"Path does not exist: {path}")
        return resolved_path
    
    def parse_skill_md(self, path: Path) -> dict:
        """
        Parse a SKILL.md file into metadata and body.

        Returns dict with: name, description, compatibility, body, path, dir
        Returns None if file doesn't match format.
        """
        content = path.read_text()

        # Match YAML frontmatter between --- markers
        match = re.match(r"^---\s*\n(.*?)\n---\s*\n(.*)$", content, re.DOTALL)
        if not match:
            return None

        frontmatter, body = match.groups()

        # Parse YAML-like frontmatter (simple key: value)
        metadata = {}
        for line in frontmatter.strip().split("\n"):
            if ":" in line:
                key, value = line.split(":", 1)
                metadata[key.strip()] = value.strip().strip("\"'")

        # Require name and description
        if "name" not in metadata or "description" not in metadata:
            return None

        return {
            "name": metadata["name"],
            "description": metadata["description"],
            "compatibility": metadata.get("compatibility"),
            "body": body.strip(),
            "path": path,
            "dir": path.parent,
        }

    def load_skills(self, skills_dir: Path):
        """
        Scan skills directory and load all valid SKILL.md files.

        Only loads metadata at startup - body is loaded on-demand.
        This keeps the initial context lean.
        """
        if not skills_dir.exists():
            return

        for skill_dir in skills_dir.iterdir():
            if not skill_dir.is_dir():
                continue

            skill_md = skill_dir / "SKILL.md"
            if not skill_md.exists():
                continue

            skill = self.parse_skill_md(skill_md)
            if skill:
                self.skills[skill["name"]] = skill

    def add_skills_by_dir(self, skills_dir: str | list[str]):
        """
        add skills from a directory
        """
        if isinstance(skills_dir, str):
            skills_dirs = [skills_dir]
        else:
            skills_dirs = skills_dir
        
        for skills_dir in skills_dirs:
            skills_dir_p = self.safe_path(skills_dir)
            self.load_skills(skills_dir=skills_dir_p)

    def get_descriptions(self) -> str:
        """
        Generate skill descriptions for system prompt.

        This is Layer 1 - only name and description, ~100 tokens per skill.
        Full content (Layer 2) is loaded only when Skill tool is called.
        """
        if not self.skills:
            return "(no skills available)"

        lines = []
        for name, skill in self.skills.items():
            line = f"- {name}: {skill['description']}"
            if skill.get("compatibility"):
                line += f" [requires: {skill['compatibility']}]"
            lines.append(line)
        return "\n".join(lines)

    def get_skill_content(self, name: str) -> str:
        """
        Get full skill content for injection.

        This is Layer 2 - the complete SKILL.md body, plus any available
        resources (Layer 3 hints).

        Returns None if skill not found.
        """
        if name not in self.skills:
            return None

        skill = self.skills[name]
        content = f"# Skill: {skill['name']}\n\n{skill['body']}"

        # List available resources (Layer 3 hints)
        resources = []
        for folder, label in [
            ("scripts", "Scripts"),
            ("references", "References"),
            ("assets", "Assets")
        ]:
            folder_path = skill["dir"] / folder
            if folder_path.exists():
                files = list(folder_path.glob("*"))
                if files:
                    resources.append(f"{label}: {', '.join(f.name for f in files)}")

        if resources:
            content += f"\n\n**Available resources in {skill['dir']}:**\n"
            content += "\n".join(f"- {r}" for r in resources)

        return content

    def list_skills(self) -> list:
        """Return list of available skill names."""
        return list(self.skills.keys())
    
    def run_skill(self, skill_name: str) -> str:
        """
        Load a skill and inject it into the conversation.

        This is the key mechanism:
        1. Get skill content (SKILL.md body + resource hints)
        2. Return it wrapped in <skill-loaded> tags
        3. Model receives this as tool_result (user message)
        4. Model now "knows" how to do the task

        Why tool_result instead of system prompt?
        - System prompt changes invalidate cache (20-50x cost increase)
        - Tool results append to end (prefix unchanged, cache hit)

        This is how production systems stay cost-efficient.
        """
        content = self.get_skill_content(skill_name)

        if content is None:
            available = ", ".join(self.list_skills()) or "none"
            return f"Error: Unknown skill '{skill_name}'. Available: {available}"

        # Wrap in tags so model knows it's skill content
        return f"""<skill-loaded name="{skill_name}">
    {content}
    </skill-loaded>

    Follow the instructions in the skill above to complete the user's task."""
        

if __name__ == "__main__":
    loader = SkillLoader(Path("/home/xiongdb/drsai_dev/examples/agent_groupchat/assistant_skill/skills"))
    # print(loader.get_descriptions())
    print(loader.list_skills())
    # print(loader.get_skill_content("pdf"))


from .drsai_model_context import DrSaiChatCompletionContext

from autogen_core.model_context import (
    ChatCompletionContext,
    ChatCompletionContextState,
    UnboundedChatCompletionContext,
    BufferedChatCompletionContext,
    TokenLimitedChatCompletionContext,
    HeadAndTailChatCompletionContext,
)

__all__ = [
    "DrSaiChatCompletionContext",
    "ChatCompletionContext",
    "ChatCompletionContextState",
    "UnboundedChatCompletionContext",
    "BufferedChatCompletionContext",
    "TokenLimitedChatCompletionContext",
    "HeadAndTailChatCompletionContext",
]





from .ragflow_memory import RAGFlowMemory, RAGFlowMemoryManager, RAGFlowMemoryConfig
from autogen_core.memory import (
    Memory, 
    MemoryContent, 
    MemoryMimeType, 
    MemoryQueryResult, 
    UpdateContextResult,
    ListMemory
)
from autogen_ext.memory.chromadb import ChromaDBVectorMemory, PersistentChromaDBVectorMemoryConfig

__all__ = [
    "RAGFlowMemory",
    "RAGFlowMemoryManager",
    "RAGFlowMemoryConfig",
    "Memory",
    "MemoryContent",
    "MemoryQueryResult",
    "UpdateContextResult",
    "MemoryMimeType",
    "ListMemory",
    "ChromaDBVectorMemory",
    "PersistentChromaDBVectorMemoryConfig",
]


from .assistant_skill import SkillAgent
from .drsai_assistant import DrSaiAssistant
from .managers.user_profile_manager import UserProfileManager
from .managers.todo_manager import TodoManager
from .managers.operater_funs import get_operator_funcs
from .managers.task_planner import TaskPlanner
from .managers.memory_manager import LongTermMemoryManager
from .managers.get_managers_tools import (
    get_agent_skills_tool,
    get_subagent_tools,
    get_todo_manager_tool,
)


from autogen_agentchat.teams._group_chat._events import (
    GroupChatAgentResponse,
    GroupChatError,
    GroupChatMessage,
    GroupChatPause,
    GroupChatRequestPublish,
    GroupChatReset,
    GroupChatResume,
    GroupChatStart,
    GroupChatTermination,
    SerializableException,
    )
from autogen_agentchat.messages import (
    BaseAgentEvent,
    BaseChatMessage,
    )
from .agent_messages import AgentLongTaskMessage, LongTaskQueryMessage
from pydantic import BaseModel
from typing import (
    List,
    Mapping,
    Sequence,
    )

class GroupChatLazyInit(BaseModel):
    
    ...

class GroupChatClose(BaseModel):
    
    ...

class GroupChatAgentLongTask(BaseModel):
    
    message: AgentLongTaskMessage | LongTaskQueryMessage
    
    


import asyncio
from abc import ABC, abstractmethod
from typing import (
    List, 
    Optional, 
    Dict, 
    Any, 
    Union, 
    Tuple,
    Mapping,
    Sequence,
    AsyncGenerator,)

from pydantic import BaseModel

from autogen_core import (
    ComponentModel,
    ComponentBase,
    Component,
    CancellationToken,
    AgentRuntime,
    MessageContext, 
    # event, 
    # rpc,
)

from autogen_agentchat.base import (
    ChatAgent, 
    Team,
    TaskResult, 
    Response,
    TerminationCondition,
    TaskRunner,
    )

from autogen_agentchat.state import BaseState

from autogen_agentchat.messages import (
    BaseAgentEvent,
    BaseChatMessage,
    AgentEvent,
    ChatMessage,
    HandoffMessage,
    MemoryQueryEvent,
    ModelClientStreamingChunkEvent,
    TextMessage,
    ToolCallExecutionEvent,
    ToolCallRequestEvent,
    ToolCallSummaryMessage,
    UserInputRequestedEvent,
    ThoughtEvent,
    StructuredMessageFactory,
    MultiModalMessage,
    Image,
    MessageFactory,
    StopMessage,
    StructuredMessage,
)

from autogen_agentchat.teams._group_chat._events import (
    GroupChatAgentResponse,
    GroupChatError,
    GroupChatMessage,
    GroupChatPause,
    GroupChatRequestPublish,
    GroupChatReset,
    GroupChatResume,
    GroupChatStart,
    GroupChatTermination,
    SerializableException,
    )
from drsai.modules.components.groupchat_running.base_running_deamon import(
    BaseRunTimeDeamon,
    GroupChatEvent,
    handle_group_chat_event,
    )
from drsai.modules.managers.database import DatabaseManager, DatabaseManagerConfig
from drsai import HepAIChatCompletionClient
from .base_group_chat_runner import BaseGroupChatRunner

class DrSaiBaseGroupChatRunnerConfig(BaseModel):
    """The configuration for the base group chat runner."""
    name: str
    max_turns: int
    model_client_config: Mapping[str, Any]
    crrent_turn: int = 0
    current_speaker: str = ""
    message_thread: List[BaseAgentEvent | BaseChatMessage] = []
    message_history: List[BaseChatMessage] = []
    database_manager: DatabaseManagerConfig|None = None
    team_config: Mapping[str, Any] = {}



class DrSaiBaseGroupChatRunner(BaseGroupChatRunner, Component[DrSaiBaseGroupChatRunnerConfig]):

    component_type = "group_chat_runner"
    component_config_schema = DrSaiBaseGroupChatRunnerConfig
    component_provider_override = "drsai.DrSaiBaseGroupChatRunner"

    def __init__(
        self, 
        name: str,
        self_id: str,
        max_turns: int,
        participant_instances: Dict[str, ChatAgent],
        output_message_queue: asyncio.Queue[BaseAgentEvent | BaseChatMessage | GroupChatTermination],
        termination_condition: TerminationCondition,
        model_client: HepAIChatCompletionClient = None,
        messages_history: List[BaseChatMessage] = [],
        sequential_message_types: List[GroupChatEvent] = [
            GroupChatStart, # handle start messages
            GroupChatAgentResponse, # handle agent responses
            GroupChatMessage, # handle GroupChat system messages
            GroupChatPause, # handle pause messages
            GroupChatResume, # handle resume messages
            GroupChatTermination, # handle termination messages
            GroupChatError, # handle error messages
            GroupChatReset, # handle reset messages
            GroupChatRequestPublish, # handle request publish messages
        ],
        emit_team_events: bool = False,
    ):
        
        self._name = name
        self._self_id = self_id
        if max_turns is not None and max_turns <= 0:
            raise ValueError("The maximum number of turns must be greater than 0.")
        self._max_turns = max_turns
        self._current_turn = 0
        self._participant_name_to_descriptions = {
            name: agent.description for name, agent in participant_instances.items()
        }
        self._output_message_queue = output_message_queue
        self._termination_condition = termination_condition
        self._model_client = model_client
        self._emit_team_events = emit_team_events
        self._messages_history = messages_history

        self._message_thread: List[BaseAgentEvent | BaseChatMessage] = []
        self._sequential_message_types = sequential_message_types

        self._current_speaker = ""
    
    @handle_group_chat_event
    async def handle_start(self, message: GroupChatStart, cancellation_token: CancellationToken | None = None) -> None:
        pass
    
    @handle_group_chat_event
    async def handle_agent_response(self, message: GroupChatAgentResponse, cancellation_token: CancellationToken | None = None) -> None:
        pass
    
    @handle_group_chat_event
    async def handle_group_chat_message(self, message: GroupChatMessage, cancellation_token: CancellationToken | None = None) -> None:
        """Handle a group chat message by appending the content to its output message queue."""
        await self._output_message_queue.put(message.message)
    
    @handle_group_chat_event
    async def handle_group_chat_error(self, message: GroupChatError, cancellation_token: CancellationToken | None = None) -> None:
        """Handle a group chat error by logging the error and signaling termination."""
        error = message.error
        termination_event = GroupChatTermination(
            message=StopMessage(content="An error occurred in the group chat.", source=self._name), error=error
        )
        await self._output_message_queue.put(termination_event)

    @handle_group_chat_event
    async def handle_reset(self, message: GroupChatReset, cancellation_token: CancellationToken | None = None) -> None:
        """Reset the group chat manager. Calling :meth:`reset` to reset the group chat manager
        and clear the message thread."""
        await self.reset()

    @handle_group_chat_event
    async def handle_pause(self, message: GroupChatPause, cancellation_token: CancellationToken | None = None) -> None:
        """Pause the group chat manager. This is a no-op in the base class."""
        pass

    @handle_group_chat_event
    async def handle_resume(self, message: GroupChatResume, cancellation_token: CancellationToken | None = None) -> None:
        """Resume the group chat manager. This is a no-op in the base class."""
        pass

    async def select_speaker(self, thread: List[BaseAgentEvent | BaseChatMessage]) -> str:
        """Select a speaker from the participants and return the
        topic type of the selected speaker."""
        pass
    
    async def publish_messages_to_participants(
            self, 
            messages: List[BaseChatMessage], 
            speaker_name: str, 
            cancellation_token: CancellationToken | None = None) -> None:
        """Select a speaker from the participants and return the
        topic type of the selected speaker."""
        pass

    async def reset(self) -> None:
        """Reset the team and all its participants to its initial state."""
        pass

    async def pause(self) -> None:
        """Pause the team and all its participants. This is useful for
        pausing the :meth:`autogen_agentchat.base.TaskRunner.run` or
        :meth:`autogen_agentchat.base.TaskRunner.run_stream` methods from
        concurrently, while keeping them alive."""
        pass

    async def resume(self) -> None:
        """Resume the team and all its participants from a pause after
        :meth:`pause` was called."""
        pass

    async def save_state(self) -> Mapping[str, Any]:
        """Save the current state of the team."""
        pass

    async def load_state(self, state: Mapping[str, Any]) -> None:
        """Load the state of the team."""
        pass

    async def validate_group_state(self, messages: List[BaseChatMessage] | None) -> None:
        """Validate the state of the group chat given the start messages.
        This is executed when the group chat manager receives a GroupChatStart event.

        Args:
            messages: A list of chat messages to validate, or None if no messages are provided.
        """
        pass

    def _to_config(self):
        pass

